#ifndef APP_WORKSPACE_H
#define APP_WORKSPACE_H

#include <memory>
#include <string>
#include <chrono>
#include <thread>
#include <hx711/common.h>
#include "user_workspace.h"
#include "keyboard.h"
#include "rfid_reader.h"
#include "screen_manager.h"
#include "spidev_lib++.h"
#include "db_driver.h"
#include "custom_hx711.h"

#include "imgui.h"
#include <map>

#define DEF_BUFF_SIZE_TINY      16
#define DEF_BUFF_SIZE_V_SMALL   32
#define DEF_BUFF_SIZE_SMALL     64
#define DEF_BUFF_SIZE           128
#define DEF_BUFF_SIZE_BIG       256
#define DEF_BUFF_SIZE_LARGE     512
#define DEF_BUFF_SIZE_HUGE      1024
#define DEF_BUFF_SIZE_EXTRA     2048

#define S2_MIN_SAMPLES_IN       50
#define S2_MAX_SAMPLES_IN       5000
#define S2_MIN_TIMEOUT_IN       1
#define S2_MAX_TIMEOUT_IN       60
#define S3_MIN_FILTER_NUMBER    0
#define S6_MIN_SAMPLES_IN       10
#define S6_MAX_SAMPLES_IN       1000

#define UNIT_SEL_CNT        3
#define S2_MEASURE_OTPS_CNT 2
#define S3_FILTER_OPT_CNT   2
#define S6_UNIT_SEL_CNT     2
#define S6_ROLE_CNT         3
#define S6_MONTH_CNT        12

#define DAY_MIN     1
#define DAY_MAX     31
#define YEAR_MIN    1900

#define HX_CONT_MAX_CONSEC_FAIL_FI 5
#define HX_CONT_MAX_CONSEC_FAIL_FM 5
#define HX_CONT_SAMPLES 5000

#define HX_TEST_CONST_IT_COUNT 5
#define HX_TEST_INC_IT_COUNT 5
#define HX_TEST_SAMP_COUNT 160
#define HX_TEST_TIME_COUNT 2

#define S5_TO_S6_CALIBRATION     0
#define S5_TO_S6_USERADD         1
#define S5_TO_S6_HX_TEST_SAMPLES 2
#define S5_TO_S6_HX_TEST_TIMEOUT 3
#define S5_TO_S6_KB_DELAY_TEST   4

#define S5_CALIBRATION_APPLIED  0
#define S5_USER_CREATED         1
#define S5_FINISHED_KB_DELAYS   2

/**
 * @brief This namespace holds structures and enums used in app_workspace class.
 * 
 */
namespace app_workspace_ns {
    /**
     * @brief Enum holding options of keyboard input mode.
     * 
     */
    enum kb_input_mode {
        NAVIGATION,
        INPUT
    };

    /**
     * @brief Enum holding options of keyboard INPUT mode types.
     * 
     */
    enum kb_input_type {
        NONE, // neutral state before correct kb_input_type is determined, should'nt be used otherwise
        INT,
        FLOAT,
        TEXT
    };

    /**
     * @brief Structure holding the values of keyboard input mode and input type.
     * 
     */
    typedef struct {
        kb_input_mode current_mode = NAVIGATION;
        kb_input_type current_type = NONE;
    } kb_input_state;

    /**
     * @brief This enum defines sizes of corresponding fonts. E.g.: SMALL_FONT is 12px.
     * 
     */
    enum font_size {
        SMALL_FONT = 12,
        NORMAL_FONT = 24,
        BIG_FONT = 48
    };

    /**
     * @brief Structure holding SPI device and SPI config which is a library structure "spi_config_t".
     * 
     */
    typedef struct {
        std::string dev;
        std::unique_ptr<spi_config_t> config = std::make_unique<spi_config_t>();
    } spi_config;
    
    /**
     * @brief Structure hodling the configuration of hx711 controller.
     * 
     */
    typedef struct {
        int data_pin;
        int clock_pin;
        int rate;
        int ref_unit;
        int offset;
    } hx711_config;
};

/**
 * @brief Structure that holds the values of the application configuration file.
 * 
 */
typedef struct {
    std::string opt_conf_path;
    std::string fonts_path;
    std::string font;
    std::string lang_path;
    std::string lang_default;
    std::string lang_current;
    std::vector<std::string> lang_options;
    bool debug_screens;
    bool log_debug;
    bool init_fail_quit;
    app_workspace_ns::spi_config keyb_conf;
    app_workspace_ns::spi_config rfid_conf;
    app_workspace_ns::hx711_config hx_conf;
    db_driver_conf db_conf;
} app_config;

/**
 * @brief One of the most importat classes in the whole project. Holds variables that define the state of the application 
 * and also provides logic for some of the variables held there. For example the hx711 controller is held and
 * used from this class.
 */
class app_workspace {
    public:
        /*************************************************************************************************************/
        /*       Properties                                                                                          */
        /*************************************************************************************************************/

        /**
         * @brief Input state holder.
         * 
         */
        std::unique_ptr<app_workspace_ns::kb_input_state> kb_in_state = 
            std::make_unique<app_workspace_ns::kb_input_state>();
        /**
         * @brief HX711 controller. This is an instance of custom_hx711 which extends library's AdvancedHX711.
         * 
         */
        std::unique_ptr<custom_hx711> hx711_controller;
        /**
         * @brief Database driver holder.
         * 
         */
        std::unique_ptr<db_driver> db_conn;

        /**
         * @brief When this is set to true, elements of current screen will be reinitialized for next frame.
         * 
         */
        bool refresh_current_screen = false;
        /**
         * @brief Flags used during text input.
         * 
         */
        bool capslock_flag = false, lshift_flag = false;

        /**
         * @brief Application config loaded from app_config.conf (main config file).
         * 
         */
        std::unique_ptr<app_config> main_config = std::make_unique<app_config>();

        /**
         * @brief Flag to indicate if a user is logged in.
         * 
         */
        bool user_logged = false;
        /**
         * @brief User's workspace. Serves similar function as app_workspace but for user data.
         * 
         */
        std::unique_ptr<user_workspace> userspace;

        /** hx711 (scale) variables */
        bool hx_measuring = false;
        bool hx_continuous = false;
        bool hx_not_finished_yet = false;
        std::chrono::time_point<std::chrono::high_resolution_clock> m_start, m_end; // m = measuring (hx - scale)
        int selected_hx_measure_opt = 0; /** 0 = samples, 1 = timeout */
        int selected_hx_result_unit = 0;
        /**
         * @brief This variable is used to observe measuring.
         * 
         */
        std::unique_ptr<measurement> observed_measurement;
         
        /**
         * @brief Map of loaded fonts. 
         * 
         */
        std::map<app_workspace_ns::font_size, ImFont*> loaded_fonts;
        /**
         * @brief Whenever a font is used, it is pushed to the font stack of ImGui library. All pushed 
         * fonts need to be popped before the end of the frame to prevent application crash.
         * 
         */
        int font_push_cnt = 0;

        /** screen_1 vars - these variables are used for gui elemnts on screen 1 **/
        char usr_name[DEF_BUFF_SIZE], usr_passwd[DEF_BUFF_SIZE_SMALL], subusr_name[DEF_BUFF_SIZE];

        /** screen_2 vars - these variables are used for gui elemnts on screen 2 **/
        const char* hx_measure_opts[S2_MEASURE_OTPS_CNT];
        int samples_in = S2_MIN_SAMPLES_IN, timeout_in = S2_MIN_TIMEOUT_IN;
        
        /** screen_3 vars - these variables are used for gui elemnts on screen 3 **/
        int selected_measuring = 0;
        const char* s3_filter_opt_labels[S3_FILTER_OPT_CNT];
        int s3_selected_filter = 0;
        unsigned long s3_filter_number = 0;

        /** screen_4 vars - these variables are used for gui elemnts on screen 4**/
        int selected_unit_option = 1; // by default values are in G
        // unit_selection MUST have an array with corresponding labels (this is currently in screen_definitions.cpp:)
        const char* unit_sel_labels[UNIT_SEL_CNT];
        const HX711::Mass::Unit unit_selection[UNIT_SEL_CNT] = 
            {HX711::Mass::Unit::MG, HX711::Mass::Unit::G, HX711::Mass::Unit::KG};

        /** screen_5 vars - these variables are used for gui elemnts on screen 5 **/
        int s5_screen_6_indicator = -1;
        int s5_success_indicator = -1;

        /** screen_6 vars - these variables are used for gui elemnts on screen 6 **/
        int s6_ref_mass_in = 0, s6_samples_in = S6_MIN_SAMPLES_IN, s6_calib_ref_unit = 0, s6_calib_offset = 0;
        int s6_selected_unit_option = 0;
        const char* s6_unit_sel_labels[S6_UNIT_SEL_CNT];
        bool s6_calibration_finished = false, s6_tare_complete = false;
        std::string s6_ref_mass_str;

        HX711::Value zero_value = 0, ref_unit = 0;

        bool s6_rfid_indicator = false; /** if available rfid is scanned for user set it to true */
        std::unique_ptr<rfid_reader::rfid_tag> s6_rfid_tag;
        int s6_selected_role = 0, s6_selected_month = 0;
        const char* s6_role_labels[S6_ROLE_CNT];
        const char* s6_month_labels[S6_MONTH_CNT];
        char s6_name_in[DEF_BUFF_SIZE] = {0}, s6_uname_in[DEF_BUFF_SIZE_SMALL] = {0};
        char s6_passwd_in[DEF_BUFF_SIZE_SMALL] = {0};
        char s6_desc_in[DEF_BUFF_SIZE_EXTRA] = {0}, s6_lastname_in[DEF_BUFF_SIZE_SMALL] = {0};
        int s6_year_in = 2022, s6_day_in = 1;



        /** Following variables are used for testing and debugging **/
        int int_test = 0;
        float float_test = 0.0f;
        char text_test[DEF_BUFF_SIZE];
        char pass_test[DEF_BUFF_SIZE_SMALL];

        int kb_delay_ms = 0, kb_delay_us = 0, kb_ev_delay_ms = 0, kb_ev_delay_us = 0;

        bool test_hx_testing = false; // true while testing in progress
        /** Time in [ms] it took to collected samples. First 5 is constant time. Second 5 is increasing timeout  */
        std::pair<long, std::string> test_hx_samples_times[HX_TEST_CONST_IT_COUNT + HX_TEST_INC_IT_COUNT];
        /** Number of collected samples in time. First 5 during constant timeout. Second 5 during increasing timeout  */
        std::pair<long, std::string> test_hx_timeout_collected[HX_TEST_CONST_IT_COUNT + HX_TEST_INC_IT_COUNT];

        /*************************************************************************************************************/
        /*       Function prototypes                                                                                 */
        /*************************************************************************************************************/

        /**
         * @brief Construct a new app workspace object
         * 
         */
        app_workspace();
        /**
         * @brief This is called after localisation dictionary is loaded. After the dict is loaded
         * arrays defined in app_workspace must be initialized so they can be used in GUI comboboxes.
         * 
         */
        void init_labels();
        /**
         * @brief Initializes the hx711 controller
         * 
         * @return int - 0 on success
         */
        int init_hx711_controller();
        /**
         * @brief Function that performs taring on the hx711 controller
         * 
         */
        void hx711_tare();
        /**
         * @brief Function that performs calibration on the hx711 controller
         * 
         */
        void hx711_calibrate();
        /**
         * @brief Performs hx711 measuring with set values of either samples or timeout. The results
         * is saved to the database and to userspace variable - measured.
         * 
         * @return int - 0 on success
         */
        int hx711_measure(); /** measures into userspace->measured based on screen2 inputs */
        /**
         * @brief Starts hx711 measuring that will run until its stopped by user.
         * 
         * @return int - 0 on success
         */
        int hx711_continuous_measure();
        /**
         * @brief Tries to open the database connection and init db_driver
         * 
         * @return int - 0 on success
         */
        int open_db_connection();
        /**
         * @brief Set the err screen and activate the screen on next frame
         * 
         * @param title of the error
         * @param label description of the error
         */
        void set_err_screen_and_activate(const char* title, const char* label);
        /**
         * @brief When this is called following gui elements will use selected font_size until next call of this or 
         * font pop.
         * 
         * @param font_size to be used
         */
        void use_font_size(app_workspace_ns::font_size font_size);
        /**
         * @brief Get the font size
         * 
         * @return app_workspace_ns::font_size enum value
         */
        app_workspace_ns::font_size get_font_size();
        /**
         * @brief Initializes font into the map
         * 
         * @param key under which font is inited
         * @param size_px size of inited font
         * @return int 0 on successs
         */
        int load_font_size(app_workspace_ns::font_size key, float size_px);
        /**
         * @brief Logs in user through scanned RFID tag. Initializes user_workspace if loading_sub is false
         * 
         * @param tag RFID tag structure
         * @param loading_sub if the user is to be logged as a subuser.
         * @return int 0 on success, otherwise if logging in cannot be done
         */
        int log_in_user_rfid(rfid_reader::rfid_tag *tag, bool loading_sub = false);
        /**
         * @brief Tests if user entered correct credentials
         * 
         * @return int 0 on success
         */
        int verify_user_cred();
        /**
         * @brief Logs in user using credentials. Credentials are gained from screen_1 buffers.
         * 
         */
        void log_in_user_cred();
        /**
         * @brief When a user is already logged in and their role is employee or admin, they can log in 
         * another user as their subuser.
         * 
         */
        void login_subuser();
        /**
         * @brief Logs out subuser
         * 
         */
        void logout_subuser();
        /**
         * @brief Logs out user and subuser if logged in
         * 
         */
        void logoff_user();
        /**
         * @brief This function validets screen 6 buffers and if everything checks out then inserts new user into the DB
         * 
         * @return int 0 on success
         */
        int register_new_user();

        /**
         * @brief Takes raw values and normalizes them using HX offset and reference unit. Resulting values are in Grams!
         * 
         * @param raw_vals values before normalization
         * @return std::vector<double>  normalized values in Grams
         */
        std::vector<double> normalize_raw_values(std::vector<HX711::Value>& raw_vals);
        /**
         * @brief Reads the HX711 sensor until @samples is collected. Also starts new thread
         * which is used to monitor ongoing measuring
         * 
         * @param samples to be measured
         * @return std::vector<HX711::Value> measured values
         */
        std::vector<HX711::Value> hx711_values_samples(int samples);
        /**
         * @brief Reads the HX711 sensor until @timeout is reached. Also starts new thread
         * which is used to monitor ongoing measuring
         * 
         * @param seconds for how long the measuring will be going
         * @return std::vector<HX711::Value> measured values
         */
        std::vector<HX711::Value> hx711_values_timeout(int seconds);

        /**
         * @brief Get the instance app_workspace which is a singleton
         * 
         * @return std::unique_ptr<app_workspace>& app_workspace pointer ref.
         */
        static std::unique_ptr<app_workspace> &get_instance();

        /**
         * @brief Get the screen manager instance. This instance is kind of singleton. Is initalized only once on startup
         * 
         * @return screen_manager* pointer
         */
        screen_manager* get_scr_mgr();
        /**
         * @brief Set the screen manager instancte. This should only be called once, when the application is starting up
         * 
         * @param scr_mgr 
         */
        void set_scr_mgr(screen_manager *scr_mgr);
        /**
         * @brief Retruns flag is 
         * 
         * @return true 
         * @return false 
         */
        bool is_debug_screens();

        /**
         * @brief Checks if user is logged in. This is determined by user_space being initialized and having
         * the user_logged flag set.
         * 
         * @return true 
         * @return false 
         */
        bool has_user();
        user_workspace* get_userspace();


        // debug / tests
        /** Prints application config on debug level */
        void print_config();
        /** Tests how long (in ms) it takes hx to take X (def 160) samples. 
         *  Does 5 iterations with X and 5 iterations with X * it */
        int test_hx_samples();
        /** Tests how many samples hx can create in X (def 2) seconds. 
         *  Does 5 iterations with X and 5 iterations with X * it */
        int test_hx_timeout();

        /** Gets if keyboard testing is happening. This is a wrapper for thread secure function */
        bool get_kb_testing();
        /** Sets keyboard testing. This is a wrapper for thread secure function */
        void set_kb_testing(bool val);

    private:
        /** This thread is used for thread that monitors measuring progress */
        std::unique_ptr<std::thread> measuring_watchdog;
        /** This thread is used for measuring */
        std::unique_ptr<std::thread> measuring_thread;
        /** This thread is used when taring and calibrating HX711. Possibly could use measuring thread instead */
        std::unique_ptr<std::thread> calib_thread;
        /** The singleton instance */
        static std::unique_ptr<app_workspace> instance;
        /** The pseudo singleton screen manager instance */
        screen_manager *scr_mgr;
        /** Font size, that is currently active */
        app_workspace_ns::font_size cur_font_size = app_workspace_ns::font_size::NORMAL_FONT;

        void clear_s1_buffers(); // zero buffers used on screen 1
        void clear_s2_buffers(); // zero buffers used on screen 2
        void clear_s6_buffers(); // zero buffers used on screen 6

};

#endif