#ifndef CUSTOM_HX711_H
#define CUSTOM_HX711_H

#include <hx711/common.h>

using namespace HX711;

/**
 * @brief custom_watcher extends HX711 library Watcher. Reason is to get access
 * to the watcher state, so the measuring can be interrupted
 * 
 */
class custom_watcher : public Watcher {
    public:
        /** Calls parent constructor */
        custom_watcher(HX711::HX711 *hx);
        
        /** Get current watcher state */
        WatchState get_watch_state();
};

/**
 * @brief Extends HX711 library AdvancedHX711. Custom implementations for 
 * value reading, so it can be observed and iterrupted.
 * 
 */
class custom_hx711 : public AdvancedHX711 {
    protected:
        custom_watcher *_wx;


    public:
        /** Calls parent constructor */
        custom_hx711(
            const int dataPin,
            const int clockPin,
            const Value refUnit = 1,
            const Value offset = 0,
            const Rate rate = Rate::HZ_10);

        /**
         * @brief getValues performs reading for @timeout. This function is overwriten to allow interruption, 
         * however is no longer used in favor of getValues that can be observed
         * 
         * @param timeout for how long reading is preformed
         * @return std::vector<Value> measured values
         */
        virtual std::vector<Value> getValues(const std::chrono::nanoseconds timeout) override;
        /**
         * @brief getValues performs reading until @samples is reached. This function is overwriten to allow 
         * interruption, however is no longer used in favor of getValues that can be observed
         * 
         * @param samples to be collected
         * @return std::vector<Value> measured values.
         */
        virtual std::vector<Value> getValues(const std::size_t samples) override;
        /**
         * @brief getValues performs reading for @timeout. This function allows observing ongoing measuring thourgh
         * @watchdong
         * 
         * @param timeout for how long reading is preformed
         * @param watchdog is a vector of values, that id periodically filled and emptied during measuring to
         *  allow measuring observation
         * @return std::vector<Value> measured values
         */
        virtual std::vector<Value> getValues(const std::chrono::nanoseconds timeout, std::vector<Value> **watchdog);
        /**
         * @brief getValues performs reading until @samples is reached. This function allows observing ongoing measuring
         *  thourgh @watchdong
         * 
         * @param timeout for how long reading is preformed
         * @param watchdog is a vector of values, that id periodically filled and emptied during measuring to
         *  allow measuring observation
         * @return std::vector<Value> measured values
         */
        virtual std::vector<Value> getValues(const std::size_t samples, std::vector<Value> **watchdog);
        /**
         * @brief When this is called watcher is paused causing reading to be stopped.
         * 
         */
        virtual void stop_reading();
};

#endif