#ifndef GUI_INPUT_H
#define GUI_INPUT_H

#include "app_workspace.h"
#include "gui_element.h"

/**
 * @brief Allowed input types
 * 
 */
enum input_type {
    IN_INT,
    IN_FLOAT,
    IN_TEXT,
    IN_PASSWORD,
    IN_PASSWORD_PLAIN,
    IN_DOUBLE
};

/**
 * @brief This is a wrapper for various ImGui input types
 * 
 */
class gui_input : public gui_element {
    private:
        app_workspace_ns::font_size font_size = app_workspace_ns::font_size::NORMAL_FONT;
        input_type type = input_type::IN_INT;
        const char *label = nullptr;
        void *data = nullptr;
        int data_size = 0;
        bool has_min_max = false;
        int i_min = INT_MIN, i_max = INT_MAX;
        float f_min = FLT_MIN, f_max = FLT_MAX;
        double d_min = DBL_MIN, d_max = DBL_MAX;
        void (*callback)() = nullptr; /** value change listener */

        /** Checks if the input has set limits and input is within bounds. If the input exceeds limits, the limit
         * value is set instead.
         */
        void check_input_limit();

    public:
        /**
         * @brief Construct a new gui input object
         * 
         * @param type input data type
         * @param data container for input data, this should correspond to chosen @type
         * @param data_size max data size, this is used mainly for text input
         * @param x if set to -1, isn't used
         * @param y if set to -1, isn't used
         * @param width if set to -1, isn't used
         * @param label input can have a label (name), nullptr when not used
         * @param font_size Default font size is NORMAL
         */
        gui_input(input_type type, void *data, int data_size, int x = -1, int y = -1, int width = -1,
                const char *label = nullptr, 
                app_workspace_ns::font_size font_size = app_workspace_ns::font_size::NORMAL_FONT);

        /** Override of parent function, that is required, because it does the actual rendering of the element */
        void render_element();
        /** Returns input type */
        input_type get_type();
        /** Returns keyboard input type, that is determined by data input type */
        app_workspace_ns::kb_input_type get_in_type();
        /** Setter for minimal and maximal values for INT inputs */
        void set_min_max(int min = INT_MIN, int max = INT_MAX);
        /** Setter for minimal and maximal values for FLOAT inputs */
        void set_min_max(float min = FLT_MIN, float max = FLT_MAX);
        /** Setter for minimal and maximal values for DOUBLE inputs */
        void set_min_max(double min = DBL_MIN, double max = DBL_MAX);
        /** Sets input value gchange listener */
        void set_value_change_action(void (*callback)());
};

#endif