#ifndef RFID_READER_H
#define RFID_READER_H

struct rfid_e;
typedef struct rfid_e rfid_event;

namespace rfid_reader {
    /**
     * @brief RFID tag structure. Stores data read from SPI.
     * 
     */
    typedef struct {
        uint16_t tag_type;
        uint8_t serial[10] = {0};
        uint8_t serial_size;
        uint8_t ser_ack[3] = {0};
        int32_t max_blocks = 64;
        uint8_t page_step = 0;
    } rfid_tag;
    

    /**
     * @brief Polls event from RFID event_queue
     * 
     * @return rfid_event* 
     */
    rfid_event* poll_event();
    /**
     * @brief Clears event queue, discarding all waiting events.
     * 
     */
    void clear_event_queue();

    /**
     * @brief RFID reader init through values.
     * 
     * @return int 
     */
    int init(const char*, uint8_t, uint32_t, uint16_t, uint8_t);
    /**
     * @brief RFID reader init from loaded configuration
     * 
     * @return int 
     */
    int init_from_conf();
    /**
     * @brief Loop that is run in its own thread, reading the RFID reader
     * 
     */
    void start_reading_cards();
    /**
     * @brief Stops the reading thread
     * 
     */
    void clean();

    // following 3 functions are taken from https://github.com/paulvha/rfid-rc522.git
    /**
     * @brief This function detects if a RFID tag was put near the reader.
     * 
     * This function was implemented by paulvha
     * https://github.com/paulvha/rfid-rc522.git
     * 
     * @return char 
     */
    char detect_tag(uint16_t *tag_type, uint8_t *buff);
    /**
     * @brief This function reads the serial number of RIFD tag put near the reader.
     * 
     * This function was implemented by paulvha
     * https://github.com/paulvha/rfid-rc522.git
     * 
     * @return char 
     */
    char read_tag_serialn(uint8_t *serial, uint8_t *serial_size, uint8_t *ser_ack, uint8_t *buff);
    /**
     * @brief This function is a loop handling the scanner. 
     * 
     * This function was implemented by paulvha and modified by trestikp (me)
     * https://github.com/paulvha/rfid-rc522.git
     * 
     * @return char 
     */
    int read_tag(rfid_tag *tag);
}

/**
 * @brief Flags of RFID events
 * 
 */
enum RFID_FLAGS_ {
    RFID_FLAGS_NONE = 0x00,
    RFID_FLAGS_REPEATING = 0x01 << 0
};

/**
 * @brief Structure used for RFID event. This structure contains read tag and flags of the event.
 * 
 */
struct rfid_e {
    rfid_reader::rfid_tag tag;
    uint8_t flags;
};

#endif