#include "config_loader.h"

#include <string.h>
#include <limits.h>
#include <errno.h>
#include <fstream>
#include <string>
#include <vector>
#include <exception>
#include <spdlog/spdlog.h>
#include "app_workspace.h"
#include "utility.h"

// for debug
#include <iostream>

#include "confini.h"

#define GEN_DEFAULT_FONTS_PATH  "fonts/"
#define GEN_DEFAULT_FONT        "09809_COURIER.ttf"
#define GEN_DEFAULT_LANG_PATH   "languages/"
#define GEN_DEFAULT_LANG        "en"
//
#define KB_DEFAULT_DEV    "/dev/spidev0.1"
#define KB_DEFAULT_MODE   0
#define KB_DEFAULT_BITS   8
#define KB_DEFAULT_SPEED  500000
#define KB_DEFAULT_DELAY  0
//
#define RFID_DEFAULT_DEV    "/dev/spidev0.0"
#define RFID_DEFAULT_MODE   0
#define RFID_DEFAULT_BITS   8
#define RFID_DEFAULT_SPEED  1000000
#define RFID_DEFAULT_DELAY  0
//
#define DB_DEFAULT_HOST     "localhost"
#define DB_DEFAULT_USER     "pi"
#define DB_DEFAULT_PASSWORD "raspberry"
#define DB_DEFAULT_DATABASE "control_system"
#define DB_DEFAULT_PORT     3306
//
#define HX_DEFAULT_DATA_PIN  25
#define HX_DEFAULT_CLOCK_PIN 18
#define HX_DEFAULT_RATE      80
#define HX_DEFAULT_REF_UNIT  -45
#define HX_DEFAULT_OFFSET    -254817


namespace config_loader {
    /** parses @str to @res and returns 0 on success, other values on error */
    int parse_int(const char* str, int *res) {
        char *tmpend;
        long val;

        errno = 0;
        val = strtol(str, &tmpend, 10);

        if (errno != 0) {// over/under flow and other error
            spdlog::error("config_loader.cpp - Int parsing error. Over/Under-flow?");
            return 1;
        }
        if (str == tmpend) {// no digits
            spdlog::error("config_loader.cpp - Int parsing error. No digits in given string!");
            return 1;
        }
        if (*tmpend != '\0') {
            spdlog::error("config_loader.cpp - Int parsing error. String contains chars other than digits!");
            return 1;
        }
        if (val < INT_MIN || val > INT_MAX) {
            spdlog::error("config_loader.cpp - Int parsing error. Parsed number is LONG but not INT.");
            return 1;
        }

        *res = val;
        return 0;
    }

    /** calls parse_int. On success dest contains parsing result on failure @default_val is put to dest
     * ** added @max_val to prevent overflows
     *  return 0 on success
     */
    int parse_and_init_int(const char* str, int *dest, int default_val, int max_val = INT_MAX) {
        if (parse_int(str, dest)) {
            spdlog::info("config_loader.cpp - Initializing int with default value.");
            *dest = default_val;
            return 1;
        } // no else needed - parse_int() already assigns result to dest

        if (*dest > max_val) {
            spdlog::info("config_loader.cpp - Initializing int with default value due to overflow.");
            *dest = default_val;
            return 1;
        }

        return 0;
    }

    /** 
     * Parses string and if parset string isn't valid (size <= 0) then inits defalt value.
     * @dest is result of parsing. @str is parsed string, @format is library data format, @default_val...
     */
    int parse_and_init_str(std::string &dest, char *str, IniFormat format, const char* default_val) {
        int len = ini_string_parse(str, format);
        if (len <= 0) {
            spdlog::info("config_loader.cpp - Initializing string with default value.");
            dest = default_val;
        } else {
            dest = str;
        }

        return 0;
    }

    /**
     * @brief Parses the actual parsing through IniDispatch. 
     * Does it have to be static? - always static in libconfini examples and doc
     * 
     * @param disp 
     * @param v_other 
     * @return int 
     */
    static int parse_and_init(IniDispatch * const disp, void * const v_other) {
        // disp->append_to = section (+subsection) name (disp->at_len = length)
        // disp->data = key (disp->d_len = length)
        // disp->value = value (disp->v_len = length)

        // see libconfini's IniNodeType for type enum
        if (disp->type == INI_UNKNOWN) {
            // FIXME - log level ?
            spdlog::error("config_loader.cpp - Couldn't determine type of config line. Key: {0}, Value: {1}",
                disp->data,
                disp->value
            );
        }

        if (disp->type == INI_KEY) {
            /* // LEAVING THIS FOR DEBUGGING
            spdlog::debug("config_loader.cpp - Encountered key. {0}, {1}",
                disp->data,
                disp->value
            );
            printf("Dispatch:\nid: %d\nat: %s\natl: %d\ndat: %s\ndatl: %d\nform: %d\ntype: %d\nval: %s\nvall: %d\n",
             disp->dispatch_id, disp->append_to, disp->at_len,
             disp->data, disp->d_len, disp->format, disp->type, disp->value, disp->v_len);
            */

            #define conf ((app_config*) v_other)
            #define IS_SECTION(SECTION) ini_array_match(SECTION, disp->append_to, '.', disp->format)
            #define IS_KEY(KEY) ini_string_match_si(KEY, disp->data, disp->format)

            if (disp->at_len == 0) {
                // read values that have no section here
            } else if (IS_SECTION("general")) { // section "general"

                if (IS_KEY("debug_screens")) {
                //if (ini_string_match_si("debug_screens", disp->data, disp->format)) {
                    // 0 = false (default value on fail) 
                    // FIXME returns int, but debug_screens is boolean - size mismatch?
                    conf->debug_screens = ini_get_bool_i(disp->value, 0, disp->format);
                } else if (IS_KEY("optional_config_path")) {
                    int len = ini_string_parse(disp->value, disp->format);
                    if (len > 0) {
                        conf->opt_conf_path = disp->value;
                    }
                } else if (IS_KEY("init_failure_quit")) {
                    conf->init_fail_quit = ini_get_bool_i(disp->value, 0, disp->format);
                } else if (IS_KEY("fonts_path")) {
                    parse_and_init_str(conf->fonts_path, disp->value, disp->format, GEN_DEFAULT_FONTS_PATH);
                    if (conf->fonts_path.back() != '/')
                        conf->fonts_path.append("/");
                    /*
                    int len = ini_string_parse(disp->value, disp->format);
                    len <= 0 ? conf->fonts_path = "fonts/" : conf->fonts_path = disp->value;
                    */
                } else if (IS_KEY("log_debug")) {
                    conf->log_debug = ini_get_bool_i(disp->value, 0, disp->format);
                } else if (IS_KEY("font")) {
                    parse_and_init_str(conf->font, disp->value, disp->format, GEN_DEFAULT_FONT);
                } else if (IS_KEY("lang_path")) {
                    parse_and_init_str(conf->lang_path, disp->value, disp->format, GEN_DEFAULT_LANG_PATH);
                    if (conf->lang_path.back() != '/')
                        conf->lang_path.append("/");
                } else if (IS_KEY("available_langs")) {
                    int size = ini_array_get_length(disp->value, ',', disp->format);
                    char *lang;
                    for (int i = 0; i < size; i++) {
                        lang = ini_array_release(&(disp->value), ',', disp->format);
                        ini_string_parse(lang, disp->format);
                        conf->lang_options.push_back(lang);
                    }
                } else if (IS_KEY("default_lang")) {
                    parse_and_init_str(conf->lang_default, disp->value, disp->format, GEN_DEFAULT_LANG);
                    conf->lang_current = conf->lang_default;
                }

            } else if (IS_SECTION("spi.keyboard")) { // section "spi.*"
                app_workspace_ns::spi_config *spi = &(conf->keyb_conf);

                if (IS_KEY("dev")) {
                    parse_and_init_str(spi->dev, disp->value, disp->format,               KB_DEFAULT_DEV);
                } else if (IS_KEY("mode")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->mode),          KB_DEFAULT_MODE, UINT8_MAX);
                } else if (IS_KEY("bits")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->bits_per_word), KB_DEFAULT_BITS, UINT8_MAX);
                } else if (IS_KEY("speed")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->speed),         KB_DEFAULT_SPEED);
                } else if (IS_KEY("delay")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->delay),         KB_DEFAULT_DELAY, UINT16_MAX);
                }

            } else if (IS_SECTION("spi.rfid")) {
                app_workspace_ns::spi_config *spi = &(conf->rfid_conf);

                if (IS_KEY("dev")) {
                    parse_and_init_str(spi->dev, disp->value, disp->format,               RFID_DEFAULT_DEV);
                } else if (IS_KEY("mode")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->mode),          RFID_DEFAULT_MODE, UINT8_MAX);
                } else if (IS_KEY("bits")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->bits_per_word), RFID_DEFAULT_BITS, UINT8_MAX);
                } else if (IS_KEY("speed")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->speed),         RFID_DEFAULT_SPEED);
                } else if (IS_KEY("delay")) {
                    parse_and_init_int(disp->value, (int*) &(spi->config->delay), RFID_DEFAULT_DELAY, UINT16_MAX);
                }

            } else if (IS_SECTION("database")) { // section "database"
                if (IS_KEY("host")) {
                    parse_and_init_str(conf->db_conf.host, disp->value, disp->format,   DB_DEFAULT_HOST);
                } else if (IS_KEY("user")) {
                    parse_and_init_str(conf->db_conf.user, disp->value, disp->format,   DB_DEFAULT_USER);
                } else if (IS_KEY("password")) {
                    parse_and_init_str(conf->db_conf.passwd, disp->value, disp->format, DB_DEFAULT_PASSWORD);
                } else if (IS_KEY("database")) {
                    parse_and_init_str(conf->db_conf.db, disp->value, disp->format,     DB_DEFAULT_DATABASE);
                } else if (IS_KEY("port")) {
                    parse_and_init_int(disp->value, (int*) &(conf->db_conf.port),       DB_DEFAULT_PORT, UINT16_MAX);
                }

            } else if (IS_SECTION("hx711")) {
                if (IS_KEY("data_pin")) {
                    parse_and_init_int(disp->value, &(conf->hx_conf.data_pin),  HX_DEFAULT_DATA_PIN);
                } else if (IS_KEY("clock_pin")) {
                    parse_and_init_int(disp->value, &(conf->hx_conf.clock_pin), HX_DEFAULT_CLOCK_PIN);
                } else if (IS_KEY("rate")) {
                    parse_and_init_int(disp->value, &(conf->hx_conf.rate),      HX_DEFAULT_RATE);
                } 
                // MOVED to opt config
                // else if (IS_KEY("ref_unit")) {
                //     parse_and_init_int(disp->value, &(conf->hx_conf.ref_unit),  HX_DEFAULT_REF_UNIT);
                // } else if (IS_KEY("offset")) {
                //     parse_and_init_int(disp->value, &(conf->hx_conf.offset),    HX_DEFAULT_OFFSET);
                // }
            }
        }

        // those settings were moved to opt config, but in case opt config is missing, they still need to be init to def
        conf->hx_conf.ref_unit = HX_DEFAULT_REF_UNIT;
        conf->hx_conf.ref_unit = HX_DEFAULT_OFFSET;

        #undef conf
        #undef IS_SECTION
        #undef IS_KEY

        return 0;
    }

    int load_main_config_ini(const char* cfg_path) {
        app_config *cfg = app_workspace::get_instance()->main_config.get();

        if (load_ini_path(
            cfg_path,
            INI_DEFAULT_FORMAT,
            NULL,
            parse_and_init,
            cfg))
        {
            spdlog::critical("config_loader.cpp - Failed to read main configuration file (app_config.conf)");
            return 1;
        }

        return 0; 
    }


    int exist_opt_config(const char* cfg_path) {
        std::fstream config(cfg_path, std::ios::in);
        if (config.is_open()) {
            config.close();
            return 0;
        } else 
            return 1;
    }

    int load_opt_config(const char* cfg_path) {
        app_config *cfg = app_workspace::get_instance()->main_config.get();
        // std::fstream config(cfg->opt_conf_path, std::ios::in);
        std::fstream config(cfg_path, std::ios::in);

        if (!config.is_open()) {
            spdlog::error("config_loader.cpp - Failed to open optional config file. Cannot read config.");
            return 1;
        }

        std::string line;
        size_t sep_index;

        while (std::getline(config, line)) {
            sep_index = line.find_first_of('#');
            if (sep_index != line.npos) { // if the line contains comment trim line only up to #
                line = line.substr(0, sep_index);
            }

            sep_index = line.find_first_of('=');
            if (sep_index == line.npos) {
                spdlog::debug("config_loader.cpp - Config line doesn't have a key! Skipping line. (Line: {0})",
                    line.c_str());
                continue;
            }

            std::string key = line.substr(0, sep_index);
            trim(key);
            std::string val = line.substr(sep_index + 1, line.length());
            trim(val);

            try {
                // warning! KEY MUST BE in lower case, not doing tolower()
                if (!key.compare(OPT_CFG_REF_UNIT_KEY)) { // is ref_unit
                    spdlog::debug("config_loader.cpp - Optional config found {0}", key.c_str());
                    cfg->hx_conf.ref_unit = std::stoi(val);
                } else if (!key.compare(OPT_CFG_OFFSET_KEY)) { // is offset
                    spdlog::debug("config_loader.cpp - Optional config found {0}", key.c_str());
                    cfg->hx_conf.offset = std::stoi(val);
                } else { // is nothing else
                    spdlog::error("config_loader.cpp - Optional config: Unknown key: {0}", key.c_str());
                }
            } catch (std::invalid_argument &ex) {
                spdlog::error("config_loader.cpp - Optional config: {0} - {1} IS NOT a number!",
                    key.c_str(), val.c_str());
            } catch (std::out_of_range &ex) {
                spdlog::error("config_loader.cpp - Optional config: {0} - {1} value is OUT OF RANGE",
                    key.c_str(), val.c_str());
            }

            spdlog::debug("config_loader.cpp - Optional config: {0} - {1}", key.c_str(), val.c_str());
        }

        config.close();
        return 0; 
    }

    int replace_opt_in_opt_config(const char* cfg_path, std::string key, std::string value) {
        std::fstream config(cfg_path, std::ios::in);
        std::vector<std::string> lines;

        if (!config.is_open()) {
            spdlog::error("config_loader.cpp - Failed to open optional config file. Cannot read config.");
            return 1;
        }

        std::string line;

        while (std::getline(config, line)) {
            lines.push_back(line);
        }
        config.close();

        std::ofstream new_config(cfg_path, std::ios::trunc);
        size_t sep_index;

        for (size_t i = 0; i < lines.size(); i++) {
            sep_index = lines.at(i).find_first_of('=');

            if (lines.at(i)[0] == '#' || sep_index == lines.at(i).npos) { // lines a comment or doesnt have key
                new_config << lines.at(i) << "\n";
                continue;
            }

            std::string ex_key = lines.at(i).substr(0, sep_index);
            trim(ex_key);

            if (!ex_key.compare(key)) {
                std::string new_line = key + "=" + value;
                new_config << new_line << "\n";
            } else {
                new_config << lines.at(i) << "\n";
            }
        }

        new_config.flush();
        new_config.close();

        return 0; 
    }
};