#include "keyboard.h"

#include <memory>
#include <SDL2/SDL.h>
#include <unistd.h>
#include <sys/time.h>
#include <vector>
#include <stdexcept>
#include <map>
#include <chrono>
#include <spdlog/spdlog.h>

#include "app_workspace.h"

#define READ_INTERVAL 20000 // in us
#define RESPONSES_AVG_CNT 200 // avarage of response times is made after this count

namespace keyboard {
/*********************************************************************************************************************/
/*       Global variables                                                                                            */
/*********************************************************************************************************************/

    // hopefully it initializes like this
    std::queue<kb_event> event_queue;
    std::mutex eq_mutex;

    // struct timeval last_read; // changed to local chrono time_point
    // std::mutex lr_mtx;
    bool kb_testing = false;
    std::mutex test_mtx;
    unsigned long responses_sum = 0, lowest_diff = 1000000, highest_diff = 0;
    int responses_passed = 0;

    std::unique_ptr<spi_config_t> spi_config;
    std::unique_ptr<SPI> my_spi = nullptr;
    bool capturing_events = true;

/*********************************************************************************************************************/
/*       Local function definitions                                                                                  */
/*********************************************************************************************************************/

    // see implementation for documentation
    uint16_t buffer_to_keycode(const uint8_t *rx_buffer);
    void push_event(uint16_t scancode, uint8_t flags);
    // std::chrono::time_point<std::chrono::high_resolution_clock>* get_last_read();

/*********************************************************************************************************************/
/*       Implementation of header function definitions                                                               */
/*********************************************************************************************************************/

    /**
     * @brief Polls event from queue into param event
     * Returns 1 on success because it is inteded to return "true" on successfull event pop
     * 
     * @param event 
     * @return int: 0 on error, 1 on success
     */
    kb_event* poll_event() {
        if (event_queue.size() <= 0)
            return NULL;

        std::lock_guard<std::mutex> eq_guard(eq_mutex);

        kb_event* res = &event_queue.front();
        event_queue.pop();

        return res;
    }

    void clear_event_queue() {
        for (size_t i = 0; i < event_queue.size(); i++) {
            event_queue.pop();
        }
    }

    /**
     * @brief Initializes spi_config and my_spi global variables and starts SPI
     * 
     * @param device 
     * @param mode 
     * @param speed 
     * @param delay 
     * @param bits_per_word 
     * @return int: 0 on success, other values on error
     */
    int init(const char* device, uint8_t mode, uint32_t speed, uint16_t delay, uint8_t bits_per_word) {
        spi_config = std::unique_ptr<spi_config_t>(new spi_config_t());
        spi_config->mode = mode;
        spi_config->speed = speed;
        spi_config->delay = delay;
        spi_config->bits_per_word = bits_per_word;

        spdlog::info("keyboard.cpp - Initializing SPI keyboard (through parameters, enable debug to print config");
        spdlog::debug("keyboard.cpp - SPI conf dev: {0}"        , device);
        spdlog::debug("keyboard.cpp - SPI conf mode: {0}"       , spi_config->mode);
        spdlog::debug("keyboard.cpp - SPI conf bits_p_w: {0}"   , spi_config->bits_per_word);
        spdlog::debug("keyboard.cpp - SPI conf speed: {0}"      , spi_config->speed);
        spdlog::debug("keyboard.cpp - SPI conf delay: {0}"      , spi_config->delay);

        my_spi = std::unique_ptr<SPI>(new SPI(device, spi_config.get()));

        if (!my_spi->begin()) {
            spdlog::critical("keyboard.cpp - Failed to start SPI communication for keyboard");
            return 1;
        }

        spdlog::info("keyboard.cpp - Successfully initialized SPI keyboard");
        return 0;
    }

    /**
     * @brief Initializes my_spi with config loaded from config file into app_workspace. Starts SPI.
     * 
     * @return int: 0 on success, other values on error
     */
    int init_from_conf() {
        app_workspace_ns::spi_config *config = &(app_workspace::get_instance()->main_config->keyb_conf);
        spdlog::info("keyboard.cpp - Initializing SPI keyboard with config (enable debug to print config)");
        spdlog::debug("keyboard.cpp - SPI conf dev: {0}"     , config->dev);
        spdlog::debug("keyboard.cpp - SPI conf mode: {0}"    , config->config->mode);
        spdlog::debug("keyboard.cpp - SPI conf bits_p_w: {0}", config->config->bits_per_word);
        spdlog::debug("keyboard.cpp - SPI conf speed: {0}"   , config->config->speed);
        spdlog::debug("keyboard.cpp - SPI conf delay: {0}"   , config->config->delay);

        my_spi = std::unique_ptr<SPI>(new SPI(config->dev.c_str(), config->config.get()));

        if (!my_spi->begin()) {
            spdlog::critical("keyboard.cpp - Failed to start SPI communication for keyboard");
            return 1;
        }

        spdlog::info("keyboard.cpp - Successfully initialized SPI keyboard");
        return 0;
    }

    /**
     * @brief  Infinite loop that reads SPI keyboard. Runs in it's own thread!
     * 
     */
    void start_capturing_events() {
        // need 4 bytes to correctly read key code (aka scan code)
        // 1st byte always == 00 (or ff) - chipselect, need 4th byte to correctly read the bytes inbetween
        uint8_t rx_buffer[4] = {0};
        uint8_t event_flags = 0;
        std::mutex tst_mtx;
        uint16_t prev_scancode = 0, cur_scancode = 0;
        std::chrono::time_point<std::chrono::high_resolution_clock> last_read;
        //struct timeval lp_time, cp_time; // lp = last_press, cp = current_press

        spdlog::info("keyboard.cpp - Starting event capturing.");

        while (capturing_events) {
            prev_scancode = cur_scancode;

            event_flags = 0;
            memset(rx_buffer, 0, 4);
            my_spi->read(rx_buffer, 4);

            cur_scancode = buffer_to_keycode(rx_buffer);

            if (prev_scancode != 0x0000 || cur_scancode != 0x0000) {
                if (cur_scancode != 0x0000)
                    event_flags |= KB_FLAGS_KEY_DOWN;
                if (cur_scancode == prev_scancode)
                    event_flags |= KB_FLAGS_REPEATING;
                if (get_kb_testing_inner()) {
                    std::chrono::time_point<std::chrono::high_resolution_clock> now = 
                        std::chrono::high_resolution_clock::now();
                    unsigned long us_diff = 
                        std::chrono::duration_cast<std::chrono::microseconds>(now - last_read).count();
                    unsigned long ms_diff = us_diff / 1000;
                    last_read = now;

                    tst_mtx.lock();
                    app_workspace::get_instance()->kb_delay_us = us_diff;
                    app_workspace::get_instance()->kb_delay_ms = ms_diff;
                    tst_mtx.unlock();

                    spdlog::info("event_keyboard.cpp - keyboard delay test: {0} ms ({1} us)", ms_diff, us_diff);

                    if (responses_passed == RESPONSES_AVG_CNT) {
                        spdlog::warn("event_keyboard.cpp - lowest delay: {0} us", lowest_diff);
                        spdlog::warn("event_keyboard.cpp - highest delay: {0} us", highest_diff);
                        spdlog::warn("event_keyboard.cpp - avarage delay over {0} samples is: {1} ms ({2} us)",
                            RESPONSES_AVG_CNT, (responses_sum / responses_passed) / 1000,
                            responses_sum / responses_passed);
                        responses_passed = 0;
                        responses_sum = 0;
                        highest_diff = 0;
                        lowest_diff = 1000000;
                    }

                    responses_passed++;
                    responses_sum += us_diff;

                    if (us_diff > highest_diff) highest_diff = us_diff;
                    if (us_diff < lowest_diff) lowest_diff = us_diff;
                    
                    event_flags |= KB_FLAGS_TEST;
                }
                
                if (cur_scancode == 0x0000) // if current_scancode is 0, send prev event as KEY UP
                    push_event(prev_scancode, event_flags);
                else
                    push_event(cur_scancode, event_flags);
            }

            usleep(READ_INTERVAL); // sleep between reads
        }
    }

    /**
     * @brief Stops event capturing and closes SPI.
     * 
     */
    void clean() {
        spdlog::info("keyboard.cpp - Cleaning SPI keyboard");
        capturing_events = false;
        // my_spi is unique_ptr, should destruct automatically
    }

    /** Returns the kb_testing flag value */
    bool get_kb_testing_inner() {
        std::lock_guard<std::mutex> kb_guard(test_mtx);
        return kb_testing;
    }

    /** Sets the kb_testing flag value */
    void set_kb_testing_inner(bool val) {
        std::lock_guard<std::mutex> kb_guard(test_mtx);
        kb_testing = val;
    }

/*********************************************************************************************************************/
/*       Implementation of local function definitions                                                               */
/*********************************************************************************************************************/

    /**
     * @brief Concats keycode from rx_buffer into one (uint16_t) variable. 
     * rx_buffer is 4 bytes (this is assumed to always be true -> no error checking).
     * Only middle bytes (index 1 and 2) are the actual keycode (byte_0 = cs, byte_3 = 00)
     * 
     * @param rx_buffer read byte buffer
     * @return uint16_t resulting keycode as one value
     */
    uint16_t buffer_to_keycode(const uint8_t *rx_buffer) {
        uint16_t key = 0;
        key = (key | rx_buffer[1]) << 8;
        key |= rx_buffer[2];
        return key;
    }

    /**
     * @brief Creates and pushes event into event_queue.
     * 
     * @param scancode one value keycode
     * @param down boolean - true if key is being held down, false (= up) on release of the key
     */
    void push_event(uint16_t scancode, uint8_t flags) {
        std::lock_guard<std::mutex> eq_guard(eq_mutex);

        kb_event event;
        event.scancode = scancode;
        event.flags = flags;
        gettimeofday(&event.timestamp, NULL);
        event_queue.push(event);
    }

    // std::chrono::time_point<std::chrono::high_resolution_clock>* get_last_read() {
    //     std::lock_guard<std::mutex> lock(lr_mtx);
    //     return &last_read;
    // }
}