// Includes from imgui template, more detailed comments in original imgui example_sdl_opengl3
#include "imgui.h"
#include "backends/imgui_impl_sdl.h"
#include "backends/imgui_impl_opengl3.h"
#include <memory>
#include <stdio.h>
#include <SDL2/SDL.h>
#if defined(IMGUI_IMPL_OPENGL_ES2)
#include <SDL_opengles2.h>
#else
#include <SDL2/SDL_opengl.h>
#endif

// My custom includes
#include <thread>
#include <spdlog/spdlog.h>

#include "app_workspace.h"
#include "keyboard.h"
#include "screen_manager.h"
#include "event_handler.h"
#include "rfid_reader.h"
#include "config_loader.h"
#include "localisation.h"
#include "utility.h"

#include <unistd.h>

// every frame several event queues is handled. To prevent event flood, only handles MAX_EVENTS_PER_FRAME each frame
// and if handling failed consecutively for MAX_CONSEC_EVENT_HANDLE_FAILURES
// clears event queue (discarding all remaining events)
#define MAX_EVENTS_PER_FRAME 10
#define MAX_CONSEC_EVENT_HANDLE_FAILURES 3

#define DEFAULT_MAIN_CFG_PATH "./app_config.conf"

int parse_args_and_prepare(int argc, char** argv) {
    input_parser args(argc, argv);

    if (args.cmd_option_exists("-c")) { // main config file path + name
        const char* val = args.get_cmd_option("-c").c_str();
        if (config_loader::load_main_config_ini(val)) {
            spdlog::critical("main.cpp - Failed to load main config. Perhaps path {0} is wrong?", val);
            return 1;
        }
    } else {
        if (config_loader::load_main_config_ini(DEFAULT_MAIN_CFG_PATH)) {
            spdlog::critical("main.cpp - Failed to load main config through default path. Please either provide path"
                "to main configuration with -c argument or make sure {0} exits.", DEFAULT_MAIN_CFG_PATH);
            return 1;
        }
    }

    return 0;
}

int main(int argc, char** argv) {
    if (parse_args_and_prepare(argc, argv)) {
        spdlog::critical("main.cpp - Failed to parse arguments and prepare basic program settings. Exiting...");
        return 1;
    }
    
    if (app_workspace::get_instance()->main_config->log_debug)
        spdlog::set_level(spdlog::level::debug); // Set global log level to debug (if true in config)

    // optional config (hx ref_unit and offset)
    if (app_workspace::get_instance()->main_config->opt_conf_path.size() > 0) {
        if (config_loader::load_opt_config(app_workspace::get_instance()->main_config->opt_conf_path.c_str())) {
            spdlog::error("main.cpp - Failed to load optional config file");
        }
    }

    // example how to change optional config
    /*
    config_loader::replace_opt_in_opt_config(app_workspace::get_instance()->main_config->opt_conf_path.c_str(),
        "offset", std::to_string(5000));
    */

    // print_config() - prints on debug level
    app_workspace::get_instance()->print_config();
    
    if (load_default_dict()) {
        spdlog::critical("Please make sure language options are set correctly in main configuration file. Exiting...");
        return 2;
    }

    // Setup SDL
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_GAMECONTROLLER) != 0) {
        spdlog::critical("main.cpp - Failed to init SDL. Error: %s\n", SDL_GetError());
        return 3;
    }

    // Decide GL+GLSL versions, leaving this whole thing, to support development on multiple platforms
#if defined(IMGUI_IMPL_OPENGL_ES2)
    // GL ES 2.0 + GLSL 100
    const char* glsl_version = "#version 100";
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, 0);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_ES);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 0);
#elif defined(__APPLE__)
    // GL 3.2 Core + GLSL 150
    const char* glsl_version = "#version 150";
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG); // Always required on Mac
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
#else
    // GL 3.0 + GLSL 130
    const char* glsl_version = "#version 130";
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, 0);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 0);
#endif

    // Create window with graphics context
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
    SDL_WindowFlags window_flags = (SDL_WindowFlags)(SDL_WINDOW_OPENGL | SDL_WINDOW_ALLOW_HIGHDPI);
    SDL_Window* window = SDL_CreateWindow(get_localized_text("APP_WINDOW_TITLE"),
            SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 480, 640, window_flags);
    SDL_GLContext gl_context = SDL_GL_CreateContext(window);
    SDL_GL_MakeCurrent(window, gl_context);
    SDL_GL_SetSwapInterval(1); // Enable vsync (is this wanted?)

    // Setup Dear ImGui context
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;       // Enable Keyboard Controls
    // this is docking branch specific, might not need this and use the normal "master" version instead
    io.ConfigFlags |= ImGuiConfigFlags_DockingEnable;           // Enable Docking
    io.ConfigFlags |= ImGuiConfigFlags_ViewportsEnable;         // Enable Multi-Viewport / Platform Windows
    //io.ConfigViewportsNoAutoMerge = true;
    //io.ConfigViewportsNoTaskBarIcon = true;

    // Setup Dear ImGui style (light is probably desired)
    // ImGui::StyleColorsDark();
    // ImGui::StyleColorsClassic();
    ImGui::StyleColorsLight();

    // When viewports are enabled we tweak WindowRounding/WindowBg so platform windows can look identical to regular ones.
    ImGuiStyle& style = ImGui::GetStyle();
    if (io.ConfigFlags & ImGuiConfigFlags_ViewportsEnable) {
        style.WindowRounding = 0.0f;
        style.Colors[ImGuiCol_WindowBg].w = 1.0f;
    }
    style.FrameBorderSize = 1.0f;

    // Setup Platform/Renderer backends
    ImGui_ImplSDL2_InitForOpenGL(window, gl_context);
    ImGui_ImplOpenGL3_Init(glsl_version);


    // State and flag vars
    ImVec4 clear_color = ImVec4(0.45f, 0.55f, 0.60f, 1.00f);
    bool main_window_open = true;
    ImGuiWindowFlags main_window_flags = 0;
//                      ImGuiWindowFlags_NoDecoration = ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoCollapse,
    main_window_flags = ImGuiWindowFlags_NoDecoration | ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoSavedSettings |
                        ImGuiWindowFlags_NoScrollWithMouse | ImGuiWindowFlags_NoMouseInputs;
    bool done = false;
    std::unique_ptr<app_workspace> &app_wrk = app_workspace::get_instance();
    std::unique_ptr<screen_manager> scr_mgr(new screen_manager());
    int rv = 0;
    std::thread keyboard_thread, rfid_thread;

    // this MUST be here, to get access to screen_manager anywhere if needed
    app_wrk->set_scr_mgr(scr_mgr.get());

    // load desired fonts
    if (app_wrk->load_font_size(app_workspace_ns::SMALL_FONT, 12.0f)) {
        spdlog::error("Failed to lod font (this is already logged from load_font_size()");
        // return 1;
    }
    if (app_wrk->load_font_size(app_workspace_ns::NORMAL_FONT, 24.0f)) {
        spdlog::error("Failed to lod font (this is already logged from load_font_size()");
        // return 1;
    }
    if (app_wrk->load_font_size(app_workspace_ns::BIG_FONT, 48.0f)) {
        spdlog::error("Failed to lod font (this is already logged from load_font_size()");
        // return 1;
    }

    // Pre-loop calls (inits)
    rv = app_wrk->init_hx711_controller();
    if (rv) {
        spdlog::critical("main.cpp - Failed to start HX711.");
        if (app_wrk->main_config->init_fail_quit)
            return 1;
    }

    rv = keyboard::init_from_conf();
    if (rv) {
        spdlog::critical("main.cpp - SPI keyboard failed to initialize. Unless another keyboard"
                      " is connected, you won't be able to control the application.");
        if (app_wrk->main_config->init_fail_quit) return 1;
    } else 
        keyboard_thread = std::thread(keyboard::start_capturing_events);

    rv = rfid_reader::init_from_conf();
    if (rv) {
        spdlog::critical("main.cpp - SPI keyboard failed to initialize. Unless another keyboard"
                      " is connected, you won't be able to control the application.");
        if (app_wrk->main_config->init_fail_quit) return 1;
    } else
        rfid_thread = std::thread(rfid_reader::start_reading_cards);

    rv = app_workspace::get_instance()->open_db_connection();
    if (rv) {
        spdlog::critical("main.cpp - Failed to open DB connection or initialize DB driver.");
        return 1; // always stop the app, the app is pointless without DB
    }

    // Main loop
    while (!done) {
        // Poll and handle events (inputs, window resize, etc.)
        // Potentially can stall the app, with event flood on rpi (slow HW)
        SDL_Event event;
        kb_event* raw_event;
        rfid_event* rfid_ev;

        while ((rfid_ev = rfid_reader::poll_event())) {
            event_handler::handle_rfid_event(rfid_ev);
        }

        while ((raw_event = keyboard::poll_event())) {
            event_handler::handle_raw_event(raw_event);
        }

        while (SDL_PollEvent(&event)) {
            // this is to see what event series occurs for example when clicking a button, good for deubg except maybe using error log level

            // handle_sdl_event = custom event handling
            if (event_handler::handle_sdl_event(&event))
            //if (event_handler::handle_event(&event, scr_mgr.get()))
                ImGui_ImplSDL2_ProcessEvent(&event);

            if (event.type == SDL_QUIT)
                done = true;
            if (event.type == SDL_WINDOWEVENT && event.window.event == SDL_WINDOWEVENT_CLOSE &&
                                                 event.window.windowID == SDL_GetWindowID(window))
                done = true;
        }

        // Start the Dear ImGui frame
        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplSDL2_NewFrame();
        ImGui::NewFrame();

        // START window data 

        bool use_work_area = true;

        const ImGuiViewport* viewport = ImGui::GetMainViewport();
        ImGui::SetNextWindowPos(use_work_area ? viewport->WorkPos : viewport->Pos);
        ImGui::SetNextWindowSize(use_work_area ? viewport->WorkSize : viewport->Size);

        ImGui::Begin(get_localized_text("FRAME_TITLE"), &main_window_open, main_window_flags);

        ImGui::PushStyleColor(ImGuiCol_NavHighlight, (ImVec4) ImColor(50, 50, 255, 255));
        ImGui::PushStyleColor(ImGuiCol_FrameBg, (ImVec4) ImColor(230, 230, 230, 255));
        ImGui::PushStyleColor(ImGuiCol_Border, (ImVec4) ImColor(50, 50, 50, 255));

        // render scren
        app_wrk->get_scr_mgr()->render_active_screen();
        //scr_mgr->render_active_screen();

        while (app_wrk->font_push_cnt > 0) {
            ImGui::PopFont();
            app_wrk->font_push_cnt--;
        }

        ImGui::PopStyleColor(3); // par must match number of PushStyleColor

        ImGui::End(); 
        // END window data

        // Rendering
        ImGui::Render();
        glViewport(0, 0, (int)io.DisplaySize.x, (int)io.DisplaySize.y);
        glClearColor(clear_color.x * clear_color.w, clear_color.y * clear_color.w, clear_color.z * clear_color.w, clear_color.w);
        glClear(GL_COLOR_BUFFER_BIT);
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Update and Render additional Platform Windows
        if (io.ConfigFlags & ImGuiConfigFlags_ViewportsEnable) {
            SDL_Window* backup_current_window = SDL_GL_GetCurrentWindow();
            SDL_GLContext backup_current_context = SDL_GL_GetCurrentContext();
            ImGui::UpdatePlatformWindows();
            ImGui::RenderPlatformWindowsDefault();
            SDL_GL_MakeCurrent(backup_current_window, backup_current_context);
        }

        SDL_GL_SwapWindow(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplSDL2_Shutdown();
    ImGui::DestroyContext();

    SDL_GL_DeleteContext(gl_context);
    SDL_DestroyWindow(window);
    SDL_Quit();

    keyboard::clean(); // stops thread if joined
    rfid_reader::clean(); // stops thread if joined

    if (keyboard_thread.joinable()) // thread exists
        keyboard_thread.detach();
    if (rfid_thread.joinable())
        rfid_thread.detach();

    // don't want to brick the application in case the thread didn't want to finish, but give them 200ms to finish
    std::this_thread::sleep_for(std::chrono::milliseconds(200));

    return 0;
}