#include "db_driver.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <exception>
#include <ctime>
#include <spdlog/spdlog.h>

#define DB_SIZE_64 64 // used for password (blob/ binary) and some strings (varchar in db)
#define DB_SIZE_128 128 // used mostly for varchar
#define DB_SIZE_2048 2049 // used for description/ large varchar buffers
#define MAX_MEDIUM_BLOB 16777216 - 1 // max medium_blob size (2^24 - 1)

#define DB_DEFAULT_HOST         "localhost"
#define DB_DEFAULT_USER         "pi"
#define DB_DEFAULT_PASSWD       "raspberry"
#define DB_DEFAULT_DB           "control_system"
#define DB_DEFAULT_PORT         3306
#define DB_DEFAULT_SOCKET       NULL
#define DB_DEFAULT_CLIENTFLAGS  0

db_driver::db_driver(db_driver_conf *conf) {
    MYSQL *conn_tmp = mysql_init(NULL);
    // db_is_open is true by default

    if (conn_tmp) {
        conn.reset(conn_tmp);
        // FIXME ? Allow default values or enforce config values (config values must be used and correct)
        if (!mysql_real_connect(conn.get(),
                !conf->host.empty()     ? conf->host.c_str()        : DB_DEFAULT_HOST,
                !conf->user.empty()     ? conf->user.c_str()        : DB_DEFAULT_USER,
                !conf->passwd.empty()   ? conf->passwd.c_str()      : DB_DEFAULT_PASSWD,
                !conf->db.empty()       ? conf->db.c_str()          : DB_DEFAULT_DB,
                conf->port              ? conf->port                : DB_DEFAULT_PORT,
                !conf->socket.empty()   ? conf->socket.c_str()      : DB_DEFAULT_SOCKET,
                conf->clientflags       ? conf->clientflags         : DB_DEFAULT_CLIENTFLAGS))
        {
            spdlog::error("db_driver.cpp - (conf) constructor error. Failed to connect to database. Err: {0} - {1}",
                mysql_errno(conn.get()), mysql_error(conn.get()));
            // mysql_close(conn.get()); // connection was never open, this also causes double free (due to unique_ptr)
            db_is_open = false;
        }
    } else {
        spdlog::error("db_driver.cpp - (conf) constructor error. Failed to init MYSQL. Returned NULL pointer.");
        // mysql_close(conn.get());
        db_is_open = false;
    }
}

db_driver::db_driver(const char* host, const char* user, const char* passwd, const char* db,
                     unsigned int port, const char* socket, unsigned long cflags)
{
    MYSQL *conn_tmp = mysql_init(0);

    if (conn_tmp) {
        conn.reset(conn_tmp);
        if (!mysql_real_connect(conn.get(), host, user, passwd, db, port, socket, cflags)) {
            spdlog::error("db_driver.cpp - (test/debug) constructor error. Failed to connect to database");
            mysql_close(conn.get());
            db_is_open = false;
        }
    } else {
        spdlog::error("db_driver.cpp - (test/debug) constructor error. Failed to init MYSQL");
        mysql_close(conn.get());
        db_is_open = false;
    }
}

bool db_driver::is_open() {
    return db_is_open;
}

int db_driver::query_user_credentials(user_cred* creds, const char* uname) {
    memset(creds->uname, 0, 64);
    memset(creds->passwd, 0, 64);
    creds->status = 0;

    int par_cnt = 1, res_cnt = 3;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql = "SELECT username, password, status FROM cs_users WHERE username = ?";
    // spdlog::debug("db_driver.cpp - fnc: query_user_credentials. SQL: {0}", sql.c_str());

    bind_par[0].buffer_type = MYSQL_TYPE_STRING;
    // WARNING! - this is potentially dangerous (casting const char* to char*), but input parameter should be written to
    bind_par[0].buffer = (char*) uname;
    bind_par[0].buffer_length = strlen(uname);

    // result data containers
    uint8_t status = 0;
    char username[DB_SIZE_64] = {0}, password[DB_SIZE_64] = {0};
    // uint8_t password[DB_SIZE_64] = {0};

    // username
    bind_res[0].buffer_type = MYSQL_TYPE_STRING;
    bind_res[0].buffer = username;
    bind_res[0].buffer_length = DB_SIZE_64;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];

    // password
    bind_res[1].buffer_type = MYSQL_TYPE_STRING; // DB type is BINARY(64) this should work
    bind_res[1].buffer = password;
    bind_res[1].buffer_length = DB_SIZE_64;
    bind_res[1].is_null = &is_null[1];
    bind_res[1].length = &lengths [1];
    bind_res[1].error = &is_error [1];

    // role
    bind_res[2].buffer_type = MYSQL_TYPE_TINY; // probably TINYINT ? (-128 to 127)
    bind_res[2].buffer = &status;
    bind_res[2].is_null = &is_null[2];
    bind_res[2].length = &lengths [2];
    bind_res[2].error = &is_error [2];

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) != 1) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Expected 1 result but got {0}!", row_cnt);
        mysql_stmt_close(stmt);
        return 2;
    }

    int rv = mysql_stmt_fetch(stmt);
    if (rv == 1 || rv == MYSQL_NO_DATA) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to fetch result even"
                      " tought it should exist. (rv: {0})", rv);
        mysql_stmt_close(stmt);
        return 1;
    }

    bool err_exit = false;
    if (is_null[0] || is_error[0] || lengths[0] <= 0) { // username
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Username mustn't be NULL, ERROR or EMPTY!");
        err_exit = true;
    } else {
        strcpy(creds->uname, username);
    }
    if (is_null[1] || is_error[1] || lengths[1] <= 0) { // password
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Password mustn't be NULL, ERROR or EMPTY!");
        err_exit = true;
    } else {
        strcpy(creds->passwd, password);
    }
    if (is_null[2] || is_error[2] || lengths[2] <= 0) { // status
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Role mustn't be NULL, ERROR or EMPTY!");
        err_exit = true;
    } else {
        creds->status = status;
    }

    mysql_stmt_close(stmt);
    if (err_exit) {
        spdlog::error("db_driver.cpp - fnc: query_user_credentials. Failed to retrieve some critical data.");
        return 1;
    }

    return 0;
}

int db_driver::query_user_data(user_cont* usr, const char* uname, uint8_t* rfid_serial, uint8_t rfid_ser_len) {
    if (uname == nullptr && rfid_serial == nullptr) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. The function must be called with either uname"
                      " or rfid serial number");
        return 1;
    }

    int par_cnt = 1, res_cnt = 13;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql;

    // rfid is prefered over uname
    if (rfid_serial) {
        sql = "SELECT id, role, username, password, rfid_serial, description, name, lastname, reg_date,"
              " last_login_date, created_at, date_of_birth, measure_count FROM cs_users WHERE rfid_serial = ?";

        unsigned long len = rfid_ser_len; // better than casting uint8_t to unsigned long
        bind_par[0].buffer_type = MYSQL_TYPE_TINY_BLOB;
        bind_par[0].buffer = rfid_serial;
        bind_par[0].length = &len;
    }

    // both is supplied, ignores uname and uses sql from rfid, however both rfid and uname shouldnt be supplied
    if (uname && !rfid_serial) {
        // WARNING! - for wildcard (%) must be added to bind value, not here, but username is expected to be exact match
        sql = "SELECT id, role, username, password, rfid_serial, description, name, lastname, reg_date,"
              " last_login_date, created_at, date_of_birth, measure_count FROM cs_users WHERE username = ?";

        bind_par[0].buffer_type = MYSQL_TYPE_STRING;
        // WARNING! - this is potentially dangerous (casting cosnt char* to char*), but input parameter should be written to
        bind_par[0].buffer = (char*) uname;
        bind_par[0].buffer_length = strlen(uname);
    }

    // spdlog::debug("db_driver.cpp - fnc: query_user_data. SQL: {0}", sql.c_str());

    // result data holders
    unsigned long id = 0, measure_count = 0; // unsigned?
    uint8_t role = 0;
    char username[DB_SIZE_64], description[DB_SIZE_2048], name[DB_SIZE_128], lastname[DB_SIZE_64];
    uint8_t rfid_ser[DB_SIZE_64], password[DB_SIZE_64]; // rfid_serial is 10B max but is stored in tinyblob (max 255B)
    MYSQL_TIME reg_date, ll_date, create_date, dob_date;

    // id
    bind_res[0].buffer_type = MYSQL_TYPE_LONG;
    bind_res[0].buffer = &id;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];

    // role
    bind_res[1].buffer_type = MYSQL_TYPE_TINY; // probably TINYINT ? (-128 to 127)
    bind_res[1].buffer = &role;
    bind_res[1].is_null = &is_null[1];
    bind_res[1].length = &lengths [1];
    bind_res[1].error = &is_error [1];

    // username
    bind_res[2].buffer_type = MYSQL_TYPE_STRING;
    bind_res[2].buffer = username;
    bind_res[2].buffer_length = DB_SIZE_64;
    bind_res[2].is_null = &is_null[2];
    bind_res[2].length = &lengths [2];
    bind_res[2].error = &is_error [2];

    // password
    bind_res[3].buffer_type = MYSQL_TYPE_STRING; // DB type is BINARY(64) this should work
    bind_res[3].buffer = password;
    bind_res[3].buffer_length = DB_SIZE_64;
    bind_res[3].is_null = &is_null[3];
    bind_res[3].length = &lengths [3];
    bind_res[3].error = &is_error [3];

    // rfid_serial
    bind_res[4].buffer_type = MYSQL_TYPE_TINY_BLOB;
    bind_res[4].buffer = rfid_ser;
    bind_res[4].buffer_length = DB_SIZE_64;
    bind_res[4].is_null = &is_null[4];
    bind_res[4].length = &lengths [4];
    bind_res[4].error = &is_error [4];

    // description
    bind_res[5].buffer_type = MYSQL_TYPE_STRING;
    bind_res[5].buffer = description;
    bind_res[5].buffer_length = DB_SIZE_2048;
    bind_res[5].is_null = &is_null[5];
    bind_res[5].length = &lengths [5];
    bind_res[5].error = &is_error [5];

    // name
    bind_res[6].buffer_type = MYSQL_TYPE_STRING;
    bind_res[6].buffer = name;
    bind_res[6].buffer_length = DB_SIZE_128;
    bind_res[6].is_null = &is_null[6];
    bind_res[6].length = &lengths [6];
    bind_res[6].error = &is_error [6];

    // lastname
    bind_res[7].buffer_type = MYSQL_TYPE_STRING;
    bind_res[7].buffer = lastname;
    bind_res[7].buffer_length = DB_SIZE_64;
    bind_res[7].is_null = &is_null[7];
    bind_res[7].length = &lengths [7];
    bind_res[7].error = &is_error [7];

    // reg_date
    bind_res[8].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[8].buffer = &reg_date;
    bind_res[8].is_null = &is_null[8];
    bind_res[8].length = &lengths [8];
    bind_res[8].error = &is_error [8];

    //last_login_date
    bind_res[9].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[9].buffer = &ll_date;
    bind_res[9].is_null = &is_null[9];
    bind_res[9].length = &lengths [9];
    bind_res[9].error = &is_error [9];

    //created_at
    bind_res[10].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[10].buffer = &create_date;
    bind_res[10].is_null = &is_null[10];
    bind_res[10].length = &lengths [10];
    bind_res[10].error = &is_error [10];

    //date_of_birth
    bind_res[11].buffer_type = MYSQL_TYPE_DATE;
    bind_res[11].buffer = &dob_date;
    bind_res[11].is_null = &is_null[11];
    bind_res[11].length = &lengths [11];
    bind_res[11].error = &is_error [11];

    // measure_count
    bind_res[12].buffer_type = MYSQL_TYPE_LONG;
    bind_res[12].buffer = &measure_count;
    bind_res[12].is_null = &is_null[12];
    bind_res[12].length = &lengths [12];
    bind_res[12].error = &is_error [12];


    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) != 1) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Expected 1 result but got {0}!", row_cnt);
        mysql_stmt_close(stmt);
        return 1;
    }

    int rv = mysql_stmt_fetch(stmt);
    if (rv == 1 || rv == MYSQL_NO_DATA) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to fetch result even"
                      " tought it should exist. (rv: {0})", rv);                          
        mysql_stmt_close(stmt);
        return 1;
    }

//  sql = "SELECT id, role, username, password, rfid_serial, description, name, lastname, reg_date,"
            //   " last_login_date, created_at, date_of_birth FROM cs_users WHERE rfid_serial = ?";

    bool err_exit = false;
    if (is_null[0] || is_error[0]) { // id
        spdlog::error("db_driver.cpp - fnc: query_user_data. ID mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        usr->id = id;
    }
    if (is_null[1] || is_error[1]) { // role
        spdlog::error("db_driver.cpp - fnc: query_user_data. Role mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        usr->role = role;
    }
    if (is_null[2] || is_error[2] || lengths[2] <= 0) { //username
        spdlog::error("db_driver.cpp - fnc: query_user_data. Username mustn't be NULL, ERROR or EMPTY!");
        err_exit = true;
    } else {
        usr->username = username;
    }
    if (is_null[3] || is_error[3]) { // password (not doing any action for password)
        spdlog::debug("db_driver.cpp - fnc: query_user_data. Failed to read password");
        err_exit = true;
    }
    if (is_error[4]) { // rfid_serial
        spdlog::error("db_driver.cpp - fnc: query_user_data. RFID is in error.");
        err_exit = true;
    } else {
        if (is_null[4] || lengths[4] <= 0) {
            spdlog::debug("db_driver.cpp - fnc: query_user_data. RFID is NULL or EMPTY");
            memset(usr->rfid_serial_bin, 0, sizeof(uint8_t) * 10);
            usr->rfid_ser_len = 0;
        } else {
            if (lengths[4] <= 10) {
                memcpy(usr->rfid_serial_bin, rfid_ser, lengths[4]);
                usr->rfid_ser_len = lengths[4];
            } else {
                spdlog::error("db_driver.cpp - fnc: query_user_data."
                              " RFID is longer than 10 Bytes which should be impossible.");
            }
        }
    }
    if (is_null[5] || is_error[5]) { // description
        spdlog::debug("db_driver.cpp - fnc: query_user_data. Description is empty or error. Not doing anything");
        // usr->description = std::string();
    } else {
        usr->description = description;
    }
    if (is_null[6] || is_error[6] || lengths[6] <= 0) { // name
        spdlog::error("db_driver.cpp - fnc: query_user_data. Name mustn't be NULL, ERROR or EMPTY!");
        err_exit = true;
    } else {
        usr->name = name;
    }
    if (is_null[7] || is_error[7] || lengths[7] <= 0) { // lastname
        spdlog::error("db_driver.cpp - fnc: query_user_data. Lastname mustn't be NULL, ERROR or EMPTY!");
        err_exit = true;
    } else {
        usr->lastname = name;
    }
    if (is_null[8] || is_error[8]) { // reg_date
        spdlog::error("db_driver.cpp - fnc: query_user_data. Reg_date mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        mysql_time_to_tm(&reg_date, &(usr->reg_date));
    }
    if (is_error[9]) { // last_login_date
        spdlog::error("db_driver.cpp - fnc: query_user_data. Last_login_date mustn't be ERROR!");
        err_exit = true;
    } else {
        if (is_null[9]) {
            spdlog::debug("db_driver.cpp - fnc: query_user_data. Last_login_date is NULL!");
        } else {
            mysql_time_to_tm(&ll_date, &(usr->last_login_date));
        }
    }
    if (is_null[10] || is_error[10]) { // created_at
        spdlog::error("db_driver.cpp - fnc: query_user_data. Created_at date mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        mysql_time_to_tm(&create_date, &(usr->created_at_date));
    }
    if (is_null[11] || is_error[11]) { // date_of_birth
        spdlog::error("db_driver.cpp - fnc: query_user_data. Date_of_birth mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        mysql_time_to_tm(&dob_date, &(usr->date_of_birth));
    }
    if (is_null[12] || is_error[12]) { // measure_count
        spdlog::error("db_driver.cpp - fnc: query_user_data. Measure_count mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        usr->measure_count = measure_count;
    }

    mysql_stmt_close(stmt);
    if (err_exit) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to retrieve some critical data.");
        return 1;
    }

    return 0;
}

int db_driver::query_username(std::string& uname, unsigned long id) {
    int par_cnt = 1, res_cnt = 1;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql = "SELECT username FROM cs_users WHERE id = ?";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = &id;

    char name_buf[DB_SIZE_64] = {0};

    // role
    bind_res[0].buffer_type = MYSQL_TYPE_STRING; // probably TINYINT ? (-128 to 127)
    bind_res[0].buffer = &name_buf;
    bind_res[0].buffer_length = DB_SIZE_64;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) != 1) {
        spdlog::error("db_driver.cpp - fnc: query_username. Got {0} results! Expected 1.", row_cnt);
        mysql_stmt_close(stmt);
        return 2;
    }

    int rv = mysql_stmt_fetch(stmt);
    if (rv == 1 || rv == MYSQL_NO_DATA) {
        spdlog::error("db_driver.cpp - fnc: query_username. Failed to fetch result even"
                      " tought it should exist. (rv: {0})", rv);
        return 1;
    }

    if (is_null[0] || is_error[0]) {
        spdlog::error("db_driver.cpp - fnc: query_username. Username mustn't be NULL or ERROR!");
        mysql_stmt_close(stmt);
        return 1;
    } else {
        uname = name_buf;
    }

    mysql_stmt_close(stmt);
    return 0;
}

int db_driver::query_measurement_headers(user_cont* usr) {
    int par_cnt = 1, res_cnt = 4;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    // if measurement was continuous, only part 1 is selected for displayed
    std::string sql = "SELECT id, measurement_number, measuring_start, measuree_id FROM"
                      " cs_measurements WHERE measuree_id = ? AND status = 1 AND measurement_part = 1"
                      " ORDER BY created_at DESC";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = (void*) &(usr->id);

    // result data containers
    unsigned long id = 0, measurement_number = 0, measuree_id = 0;
    MYSQL_TIME measuring_start;

    // id
    bind_res[0].buffer_type = MYSQL_TYPE_LONG;
    bind_res[0].buffer = &id;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];

    // measurement_number
    bind_res[1].buffer_type = MYSQL_TYPE_LONG;
    bind_res[1].buffer = &measurement_number;
    bind_res[1].is_null = &is_null[1];
    bind_res[1].length = &lengths [1];
    bind_res[1].error = &is_error [1];

    // measuring_start
    bind_res[2].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[2].buffer = &measuring_start;
    bind_res[2].is_null = &is_null[2];
    bind_res[2].length = &lengths [2];
    bind_res[2].error = &is_error [2];

    // measuree_id
    bind_res[3].buffer_type = MYSQL_TYPE_LONG;
    bind_res[3].buffer = &measuree_id;
    bind_res[3].is_null = &is_null[3];
    bind_res[3].length = &lengths [3];
    bind_res[3].error = &is_error [3];

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    //can be used to get header = measurement count
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) <= 0) {
        spdlog::info("db_driver.cpp - fnc: query_measurement_headers. No rows fetched from stored from DB."
            " Assuming that user has no measurements yet.");
        mysql_stmt_close(stmt);
        return 0;
    } else {
        spdlog::info("db_driver.cpp - fnc: query_measurement_headers. Fetched {0} headers.", row_cnt);
    }

    usr->measur_headers.clear();
    bool err_exit = false;
    while (1) {

        int rv = mysql_stmt_fetch(stmt);
        if (rv == 1 || rv == MYSQL_NO_DATA) {
            break;
        }

        measurement_header *h = (measurement_header*) calloc(1, sizeof(measurement_header));

        if (is_null[0] || is_error[0]) { // id
            spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Measurement ID mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            h->id = id;
        }
        if (is_null[1] || is_error[1]) { // measurement_number
            spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Measuring number mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            h->measurement_number = measurement_number;
        }
        if (is_null[2] || is_error[2]) { // measuring start
            spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Measuring start mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            mysql_time_to_tm(&measuring_start, &(h->measuring_start));
        }
        if (is_null[3] || is_error[3]) { // measuree_id
            spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Measuree_id mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            h->measuree_id = measuree_id;
        }

        usr->measur_headers.push_back(std::unique_ptr<measurement_header>(h));
    }

    mysql_stmt_close(stmt);
    if (err_exit) {
        spdlog::error("db_driver.cpp - fnc: query_measurement_headers. Failed to retrieve some critical data.");
        return 1;
    }

    return 0;
}

int db_driver::query_measurement(measurement* m, unsigned long p_id) {
    int par_cnt = 1, res_cnt = 17;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql = "SELECT id, measurement_number, measuree_id, measurer_id, measuring_length,"
                      " measuring_start, measuring_end, description, value_length, value_count, m_values, x_val_median,"
                      " x_val_avg, all_val_median, all_val_avg, created_at, measurement_part"
                      " FROM cs_measurements WHERE id = ?";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = (void*) &p_id;

    unsigned long id = 0, measurement_number = 0, measuree_id = 0, measurer_id = 0, measuring_length = 0;
    unsigned long value_count = 0, measurement_part = 0;
    char description[DB_SIZE_2048] = {0};
    uint8_t value_length = 0;
    MYSQL_TIME measuring_start, measuring_end, created_at;
    double *m_values = (double*) malloc(MAX_MEDIUM_BLOB); // using malloc to save a few operations
    double x_val_med = 0.0, x_val_avg = 0.0, all_val_med = 0.0, all_val_avg = 0.0;

    if (m_values == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to allocated buffer for values");
        return 1;
    } else {
        memset(m_values, 0, MAX_MEDIUM_BLOB);
    }

    // id
    bind_res[0].buffer_type = MYSQL_TYPE_LONG;
    bind_res[0].buffer = &id;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];
    
    // measurement_number
    bind_res[1].buffer_type = MYSQL_TYPE_LONG;
    bind_res[1].buffer = &measurement_number;
    bind_res[1].is_null = &is_null[1];
    bind_res[1].length = &lengths [1];
    bind_res[1].error = &is_error [1];

    // measuree_id
    bind_res[2].buffer_type = MYSQL_TYPE_LONG;
    bind_res[2].buffer = &measuree_id;
    bind_res[2].is_null = &is_null[2];
    bind_res[2].length = &lengths [2];
    bind_res[2].error = &is_error [2];

    // measurer_id
    bind_res[3].buffer_type = MYSQL_TYPE_LONG;
    bind_res[3].buffer = &measurer_id;
    bind_res[3].is_null = &is_null[3];
    bind_res[3].length = &lengths [3];
    bind_res[3].error = &is_error [3];

    // measuring_length
    bind_res[4].buffer_type = MYSQL_TYPE_LONG;
    bind_res[4].buffer = &measuring_length;
    bind_res[4].is_null = &is_null[4];
    bind_res[4].length = &lengths [4];
    bind_res[4].error = &is_error [4];

    // measuring_start
    bind_res[5].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[5].buffer = &measuring_start;
    bind_res[5].is_null = &is_null[5];
    bind_res[5].length = &lengths [5];
    bind_res[5].error = &is_error [5];

    // measuring end
    bind_res[6].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[6].buffer = &measuring_end;
    bind_res[6].is_null = &is_null[6];
    bind_res[6].length = &lengths [6];
    bind_res[6].error = &is_error [6];

    // description
    bind_res[7].buffer_type = MYSQL_TYPE_STRING;
    bind_res[7].buffer = description;
    bind_res[7].buffer_length = DB_SIZE_2048;
    bind_res[7].is_null = &is_null[7];
    bind_res[7].length = &lengths [7];
    bind_res[7].error = &is_error [7];

    // value_length
    bind_res[8].buffer_type = MYSQL_TYPE_TINY; // probably TINYINT ? (-128 to 127)
    bind_res[8].buffer = &value_length;
    bind_res[8].is_null = &is_null[8];
    bind_res[8].length = &lengths [8];
    bind_res[8].error = &is_error [8];

    // value_count
    bind_res[9].buffer_type = MYSQL_TYPE_LONG;
    bind_res[9].buffer = &value_count;
    bind_res[9].is_null = &is_null[9];
    bind_res[9].length = &lengths [9];
    bind_res[9].error = &is_error [9];

    // m_values
    bind_res[10].buffer_type = MYSQL_TYPE_MEDIUM_BLOB;
    bind_res[10].buffer = m_values;
    bind_res[10].buffer_length = MAX_MEDIUM_BLOB;
    bind_res[10].is_null = &is_null[10];
    bind_res[10].length = &lengths [10];
    bind_res[10].error = &is_error [10];

    // x_val_median
    bind_res[11].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[11].buffer = &x_val_med;
    bind_res[11].is_null = &is_null[11];
    bind_res[11].length = &lengths [11];
    bind_res[11].error = &is_error [11];

    // x_val_avg
    bind_res[12].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[12].buffer = &x_val_avg;
    bind_res[12].is_null = &is_null[12];
    bind_res[12].length = &lengths [12];
    bind_res[12].error = &is_error [12];

    // all_val_med
    bind_res[13].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[13].buffer = &all_val_med;
    bind_res[13].is_null = &is_null[13];
    bind_res[13].length = &lengths [13];
    bind_res[13].error = &is_error [13];

    // all_val_avg
    bind_res[14].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[14].buffer = &all_val_avg;
    bind_res[14].is_null = &is_null[14];
    bind_res[14].length = &lengths [14];
    bind_res[14].error = &is_error [14];

    // created at
    bind_res[15].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[15].buffer = &created_at;
    bind_res[15].is_null = &is_null[15];
    bind_res[15].length = &lengths [15];
    bind_res[15].error = &is_error [15];

    // measurement_part
    bind_res[16].buffer_type = MYSQL_TYPE_LONG;
    bind_res[16].buffer = &measurement_part;
    bind_res[16].is_null = &is_null[16];
    bind_res[16].length = &lengths [16];
    bind_res[16].error = &is_error [16];


    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to init prepared statement.");
        free(m_values);
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        free(m_values);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        free(m_values);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        free(m_values);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        free(m_values);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        free(m_values);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) != 1) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Expected 1 result but got {0}!", row_cnt);
        mysql_stmt_close(stmt);
        free(m_values);
        return 2;
    }

    int rv = mysql_stmt_fetch(stmt);
    if (rv == 1 || rv == MYSQL_NO_DATA) {
        spdlog::error("db_driver.cpp - fnc: query_measurement. Failed to fetch result even"
                      " tought it should exist. (rv: {0})", rv);
        mysql_stmt_close(stmt);
        free(m_values);
        return 1;
    }

    bool err_exit = false;
    if (is_null[0] || is_error[0]) { // id
        spdlog::error("db_driver.cpp - fnc: query_user_data. ID mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->id = id;
    }
    if (is_null[1] || is_error[1]) { // measurement_number
        spdlog::error("db_driver.cpp - fnc: query_user_data. measurement_number mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->measurement_number = measurement_number;
    }
    if (is_null[2] || is_error[2]) { // measuree_id
        spdlog::error("db_driver.cpp - fnc: query_user_data. measuree_id mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->measuree_id = measuree_id;
    }
    if (is_null[3] || is_error[3]) { // measurer_id
        spdlog::error("db_driver.cpp - fnc: query_user_data. measurer_id mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->measurer_id = measurer_id;
    }
    if (is_null[4] || is_error[4]) { // measuring_length
        spdlog::debug("db_driver.cpp - fnc: query_user_data. measuring_length is NULL or ERROR!");
    } else {
        m->measuring_length = measuring_length;
    }
    if (is_null[5] || is_error[5]) { // measuring_start
        spdlog::error("db_driver.cpp - fnc: query_user_data. measuring_start mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        mysql_time_to_tm(&measuring_start, &(m->measuring_start));
    }
    if (is_null[6] || is_error[6]) { // measuring_end
        spdlog::error("db_driver.cpp - fnc: query_user_data. measuring_end mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        mysql_time_to_tm(&measuring_end, &(m->measuring_end));
    }
    if (is_null[7] || is_error[7]) { // description
        spdlog::debug("db_driver.cpp - fnc: query_user_data. description is NULL or ERROR");
        m->description = "";
    } else {
        m->description = description;
    }

    // currently values are in double
    m->value_length = sizeof(double);

    // if (is_null[8] || is_error[8]) { // value_length
    //     spdlog::debug("db_driver.cpp - fnc: query_user_data. value_length is NULL or ERROR!");
    //     m->value_length = sizeof(double);
    // } else {
    //     if (value_length <= 0) {
    //         // value_length must be positive number
    //         m->value_length = sizeof(double);
    //     } else {
    //         m->value_length = value_length;
    //     }
    // }
    if (is_null[9] || is_error[9]) { // value_count
        spdlog::debug("db_driver.cpp - fnc: query_user_data. value_count is NULL or ERROR!");
        m->value_count = lengths[10] / m->value_length; // lengths[10] is m_value length (length of bin data)
    } else {
        m->value_count = value_count;
    }
    if (is_null[10] || is_error[10]) { // m_values
        spdlog::info("db_driver.cpp - fnc: query_user_data. m_values is NULL or ERROR!"
            " This is accaptable, but shouldn't occur");
    } else {
        // spdlog::debug("values length: {0}", lengths[10]);
        // m->value_length is init sooner in this function. (If not read from db should be size of double), lengths is overall Byte size
        unsigned long val_cnt = lengths[10] / m->value_length;
        for (unsigned long i = 0; i < val_cnt; i++) {
            m->values.push_back(m_values[i]);
        }
    }
    if (is_null[11] || is_error[11]) { // x_val_median
        spdlog::error("db_driver.cpp - fnc: query_user_data. x_val_median mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->x_val_med = x_val_med;
    }
    if (is_null[12] || is_error[12]) { // x_val_avg
        spdlog::error("db_driver.cpp - fnc: query_user_data. x_val_avg mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->x_val_avg = x_val_avg;
    }
    if (is_null[13] || is_error[13]) { // all_val_median
        spdlog::error("db_driver.cpp - fnc: query_user_data. all_val_median mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->all_val_med = all_val_med;
    }
    if (is_null[14] || is_error[14]) { // all_val_avg
        spdlog::error("db_driver.cpp - fnc: query_user_data. all_val_avg mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->all_val_avg = all_val_avg;
    }
    if (is_null[15] || is_error[15]) { // created_at
        spdlog::debug("db_driver.cpp - fnc: query_user_data. measuring_end is NULL or ERROR!");
    } else {
        mysql_time_to_tm(&created_at, &(m->created_at));
    }
    if (is_null[16] || is_error[16]) { // measurement_part
        spdlog::error("db_driver.cpp - fnc: query_user_data. measurement_part mustn't be NULL or ERROR!");
        err_exit = true;
    } else {
        m->measurement_part = measurement_part;
    }

    free(m_values);
    if (err_exit) {
        spdlog::error("db_driver.cpp - fnc: query_user_data. Failed to retrieve some critical data.");
        return 1;
    }

    return 0;
}

int db_driver::query_continuous_measurement(std::vector<measurement> &ms, unsigned long m_number, 
    unsigned long p_measuree_id)
{
    int par_cnt = 2, res_cnt = 13;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql = "SELECT id, measurement_number, measuree_id, measurer_id, measuring_length,"
                      " measuring_start, measuring_end, value_count, x_val_median,"
                      " x_val_avg, all_val_median, all_val_avg, measurement_part"
                      " FROM cs_measurements WHERE measurement_number = ? AND measuree_id = ?";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = (void*) &m_number;

    bind_par[1].buffer_type = MYSQL_TYPE_LONG;
    bind_par[1].buffer = (void*) &p_measuree_id;

    unsigned long id = 0, measurement_number = 0, measuree_id = 0, measurer_id = 0, measuring_length = 0;
    unsigned long value_count = 0, measurement_part = 0;
    MYSQL_TIME measuring_start, measuring_end;
    double x_val_med = 0.0, x_val_avg = 0.0, all_val_med = 0.0, all_val_avg = 0.0;

    // id
    bind_res[0].buffer_type = MYSQL_TYPE_LONG;
    bind_res[0].buffer = &id;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];
    
    // measurement_number
    bind_res[1].buffer_type = MYSQL_TYPE_LONG;
    bind_res[1].buffer = &measurement_number;
    bind_res[1].is_null = &is_null[1];
    bind_res[1].length = &lengths [1];
    bind_res[1].error = &is_error [1];

    // measuree_id
    bind_res[2].buffer_type = MYSQL_TYPE_LONG;
    bind_res[2].buffer = &measuree_id;
    bind_res[2].is_null = &is_null[2];
    bind_res[2].length = &lengths [2];
    bind_res[2].error = &is_error [2];

    // measurer_id
    bind_res[3].buffer_type = MYSQL_TYPE_LONG;
    bind_res[3].buffer = &measurer_id;
    bind_res[3].is_null = &is_null[3];
    bind_res[3].length = &lengths [3];
    bind_res[3].error = &is_error [3];

    // measuring_length
    bind_res[4].buffer_type = MYSQL_TYPE_LONG;
    bind_res[4].buffer = &measuring_length;
    bind_res[4].is_null = &is_null[4];
    bind_res[4].length = &lengths [4];
    bind_res[4].error = &is_error [4];

    // measuring_start
    bind_res[5].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[5].buffer = &measuring_start;
    bind_res[5].is_null = &is_null[5];
    bind_res[5].length = &lengths [5];
    bind_res[5].error = &is_error [5];

    // measuring end
    bind_res[6].buffer_type = MYSQL_TYPE_TIMESTAMP;
    bind_res[6].buffer = &measuring_end;
    bind_res[6].is_null = &is_null[6];
    bind_res[6].length = &lengths [6];
    bind_res[6].error = &is_error [6];

    // value_count
    bind_res[7].buffer_type = MYSQL_TYPE_LONG;
    bind_res[7].buffer = &value_count;
    bind_res[7].is_null = &is_null[7];
    bind_res[7].length = &lengths [7];
    bind_res[7].error = &is_error [7];

    // x_val_median
    bind_res[8].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[8].buffer = &x_val_med;
    bind_res[8].is_null = &is_null[8];
    bind_res[8].length = &lengths [8];
    bind_res[8].error = &is_error [8];

    // x_val_avg
    bind_res[9].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[9].buffer = &x_val_avg;
    bind_res[9].is_null = &is_null[9];
    bind_res[9].length = &lengths [9];
    bind_res[9].error = &is_error [9];

    // all_val_med
    bind_res[10].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[10].buffer = &all_val_med;
    bind_res[10].is_null = &is_null[10];
    bind_res[10].length = &lengths [10];
    bind_res[10].error = &is_error [10];

    // all_val_avg
    bind_res[11].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_res[11].buffer = &all_val_avg;
    bind_res[11].is_null = &is_null[11];
    bind_res[11].length = &lengths [11];
    bind_res[11].error = &is_error [11];

    // measurement_part
    bind_res[12].buffer_type = MYSQL_TYPE_LONG;
    bind_res[12].buffer = &measurement_part;
    bind_res[12].is_null = &is_null[12];
    bind_res[12].length = &lengths [12];
    bind_res[12].error = &is_error [12];


    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement."
            " Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }

    bool err_exit = false;

    while (1) {
        int res = mysql_stmt_fetch(stmt);
        if (res == 1 || res == MYSQL_NO_DATA) {
            spdlog::debug("breaking with {0}", res);
            break;
        }

        measurement m;

        if (is_null[0] || is_error[0]) { // id
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. ID mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.id = id;
        }
        if (is_null[1] || is_error[1]) { // measurement_number
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. "
                "measurement_number mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.measurement_number = measurement_number;
        }
        if (is_null[2] || is_error[2]) { // measuree_id
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. measuree_id mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.measuree_id = measuree_id;
        }
        if (is_null[3] || is_error[3]) { // measurer_id
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. measurer_id mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.measurer_id = measurer_id;
        }
        if (is_null[4] || is_error[4]) { // measuring_length
            spdlog::debug("db_driver.cpp - fnc: query_continuous_measurement. measuring_length is NULL or ERROR!");
        } else {
            m.measuring_length = measuring_length;
        }
        if (is_null[5] || is_error[5]) { // measuring_start
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. "
                "measuring_start mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            mysql_time_to_tm(&measuring_start, &(m.measuring_start));
        }
        if (is_null[6] || is_error[6]) { // measuring_end
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. measuring_end mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            mysql_time_to_tm(&measuring_end, &(m.measuring_end));
        }

        // currently values are in double (save select from db with this)
        m.value_length = sizeof(double);

        if (is_null[7] || is_error[7]) { // value_count
            spdlog::debug("db_driver.cpp - fnc: query_continuous_measurement. value_count is NULL or "
                "ERROR but that isn't important in continuous measurement.");
        } else {
            m.value_count = value_count;
        }
        if (is_null[8] || is_error[8]) { // x_val_median
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. x_val_median mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.x_val_med = x_val_med;
        }
        if (is_null[9] || is_error[9]) { // x_val_avg
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. x_val_avg mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.x_val_avg = x_val_avg;
        }
        if (is_null[10] || is_error[10]) { // all_val_median
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. all_val_median mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.all_val_med = all_val_med;
        }
        if (is_null[11] || is_error[11]) { // all_val_avg
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. all_val_avg mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.all_val_avg = all_val_avg;
        }
        if (is_null[12] || is_error[12]) { // measurement_part
            spdlog::error("db_driver.cpp - fnc: query_continuous_measurement."
                " measurement_part mustn't be NULL or ERROR!");
            err_exit = true;
        } else {
            m.measurement_part = measurement_part;
        }

        ms.push_back(m);
    }

    if (err_exit) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to retrieve some critical data.");
        return 1;
    }

    return 0;
}

int db_driver::is_measurement_continuous(unsigned long m_number, unsigned long measuree_id) {
    int par_cnt = 2, res_cnt = 1;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql = "SELECT COUNT(id) FROM cs_measurements WHERE measurement_number = ? AND measuree_id = ?";

    // m_number
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = &m_number;

    // measuree_id
    bind_par[1].buffer_type = MYSQL_TYPE_LONG;
    bind_par[1].buffer = &measuree_id;

    long m_count = 0;

    // measurement count
    bind_res[0].buffer_type = MYSQL_TYPE_LONG;
    bind_res[0].buffer = &m_count;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. "
            "Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) != 1) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Got {0} results! Expected 1.", row_cnt);
        mysql_stmt_close(stmt);
        return 2;
    }

    int rv = mysql_stmt_fetch(stmt);
    if (rv == 1 || rv == MYSQL_NO_DATA) {
        spdlog::error("db_driver.cpp - fnc: query_continuous_measurement. Failed to fetch result even"
                      " tought it should exist. (rv: {0})", rv);
        return 1;
    }

    if (is_null[0] || is_error[0]) {
        spdlog::error("db_driver.cpp - fnc: . Count mustn't be NULL or ERROR!");
        mysql_stmt_close(stmt);
        return 1;
    }

    mysql_stmt_close(stmt);

    // 1 if continuous, 9 if single, 1 otherwise (shouldn't occur)
    if (m_count > 1) 
        return 0;
    if (m_count == 1) 
        return 9;
    else
        return 1;
}

int db_driver::is_username_available(const char* uname) {
    int par_cnt = 1;
    MYSQL_BIND bind_par[par_cnt];

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);

    std::string sql = "SELECT id FROM cs_users WHERE username = ?";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_STRING;
    bind_par[0].buffer = (char*) uname;
    bind_par[0].buffer_length = strlen(uname);

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: is_username_available. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: is_username_available. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: is_username_available. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: is_username_available. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: is_username_available. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) > 1) {
        spdlog::error("db_driver.cpp - fnc: is_username_available. Got {0} results!"
            "This shouldn't occur, because username should be unique!!", row_cnt);
        mysql_stmt_close(stmt);
        return 2;
    }
    if (row_cnt == 1) {
        spdlog::info("db_driver.cpp - fnc: is_username_available. Username already exists!");
        mysql_stmt_close(stmt);
        return 3;
    }

    spdlog::info("db_driver.cpp - fnc: is_username_available. Username is available.");
    mysql_stmt_close(stmt);
    return 0;
}

int db_driver::is_rfid_serial_available(uint8_t *serial, uint8_t ser_len) {
    int par_cnt = 1;
    MYSQL_BIND bind_par[par_cnt];

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);

    std::string sql = "SELECT id FROM cs_users WHERE rfid_serial = ?"; // maybe add status = 1?

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_TINY_BLOB;
    bind_par[0].buffer = serial;
    bind_par[0].buffer_length = ser_len;

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: is_rfid_serial_available. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: is_rfid_serial_available. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: is_rfid_serial_available. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: is_rfid_serial_available. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: is_rfid_serial_available. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) > 1) {
        spdlog::error("db_driver.cpp - fnc: is_rfid_serial_available. Got {0} results!"
            "This shouldn't occur, because username should be unique!!", row_cnt);
        mysql_stmt_close(stmt);
        return 2;
    }
    if (row_cnt == 1) {
        spdlog::info("db_driver.cpp - fnc: is_rfid_serial_available. RFID already exists!");
        mysql_stmt_close(stmt);
        return 3;
    }

    spdlog::info("db_driver.cpp - fnc: is_rfid_serial_available. RFID is available.");
    mysql_stmt_close(stmt);
    return 0;
}

int db_driver::increment_user_measurement_count(unsigned long id) {
    int par_cnt = 1;
    MYSQL_BIND bind_par[par_cnt];

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);

    std::string sql = "UPDATE cs_users SET measure_count = measure_count + 1 WHERE id = ?";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = (void*) &id;

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: increment_user_measurement_count. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: increment_user_measurement_count. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: increment_user_measurement_count."
            " Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: increment_user_measurement_count. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }

    mysql_stmt_close(stmt);
    return 0;
}

int db_driver::can_login_subuser(const char* uname) {
    int par_cnt = 1, res_cnt = 1;
    MYSQL_BIND bind_par[par_cnt], bind_res[res_cnt];
    unsigned long lengths [res_cnt] = {0};
    my_bool is_null       [res_cnt] = {0};
    my_bool is_error      [res_cnt] = {0};

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * res_cnt);

    std::string sql = "SELECT role FROM cs_users WHERE username = ?";

    // user id param
    bind_par[0].buffer_type = MYSQL_TYPE_STRING;
    bind_par[0].buffer = (char*) uname;
    bind_par[0].buffer_length = strlen(uname);

    uint8_t role;

    // role
    bind_res[0].buffer_type = MYSQL_TYPE_TINY; // probably TINYINT ? (-128 to 127)
    bind_res[0].buffer = &role;
    bind_res[0].is_null = &is_null[0];
    bind_res[0].length = &lengths [0];
    bind_res[0].error = &is_error [0];

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to bind results to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to store result of prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    unsigned long long row_cnt = 0;
    if ((row_cnt = mysql_stmt_num_rows(stmt)) != 1) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Got {0} results! Expected 1.", row_cnt);
        mysql_stmt_close(stmt);
        return 2;
    }

    int rv = mysql_stmt_fetch(stmt);
    if (rv == 1 || rv == MYSQL_NO_DATA) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Failed to fetch result even"
                      " tought it should exist. (rv: {0})", rv);
        return 1;
    }

    mysql_stmt_close(stmt);

    if (is_null[0] || is_error[0]) {
        spdlog::error("db_driver.cpp - fnc: can_login_subuser. Role mustn't be NULL or ERROR!");
        return 1;
    }

    return role == 0 ? 0 : 3; // only allow patients (role = 0)
}

int db_driver::update_user_rfid(unsigned long user_id, uint8_t* serial, uint8_t ser_len) {
    int par_cnt = 2;
    MYSQL_BIND bind_par[par_cnt];
    unsigned long serial_len = ser_len;

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);

    std::string sql = "UPDATE cs_users SET rfid_serial = ? WHERE id = ?";

    // rfid_serial value
    bind_par[0].buffer_type = MYSQL_TYPE_TINY_BLOB;
    bind_par[0].buffer = serial;
    bind_par[0].length = &serial_len;

    // user id param
    bind_par[1].buffer_type = MYSQL_TYPE_LONG;
    bind_par[1].buffer = (void*) &(user_id);

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: update_user_rfid. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: update_user_rfid. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: update_user_rfid. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: update_user_rfid. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }

    mysql_stmt_close(stmt);
    return 0;
}

int db_driver::insert_user(user_cont *new_usr, const char* password) {
    if (password == nullptr || strlen(password) <= 0) {
        spdlog::error("db_driver.cpp - fnc: insert_user. Password musn't be null or empty");
        return 2;
    }
    
    int par_cnt = 8;
    MYSQL_BIND bind_par[par_cnt];

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);

    std::string sql = "INSERT INTO"
                      " cs_users(role, username, password, rfid_serial, description, name, lastname, date_of_birth)"
                      " VALUES (?, ?, ?, ?, ?, ?, ?, ?)";

    // role
    bind_par[0].buffer_type = MYSQL_TYPE_TINY; // probably TINYINT ? (-128 to 127)
    bind_par[0].buffer = &(new_usr->role);

    // username
    bind_par[1].buffer_type = MYSQL_TYPE_STRING;
    bind_par[1].buffer = (char*) new_usr->username.c_str();
    bind_par[1].buffer_length = new_usr->username.size() + 1;

    // password
    bind_par[2].buffer_type = MYSQL_TYPE_STRING; // DB type is BINARY(64) this should work
    bind_par[2].buffer = (char*) password;
    bind_par[2].buffer_length = strlen(password);

    // rfid_serial
    bind_par[3].buffer_type = MYSQL_TYPE_TINY_BLOB;
    if (new_usr->rfid_ser_len <= 0) {
        bind_par[3].buffer = nullptr;
    } else {
        bind_par[3].buffer = new_usr->rfid_serial_bin;
    }
    bind_par[3].buffer_length = new_usr->rfid_ser_len;

    // description
    bind_par[4].buffer_type = MYSQL_TYPE_STRING;
    bind_par[4].buffer = (char*) new_usr->description.c_str();
    bind_par[4].buffer_length = new_usr->description.size() + 1;

    // name
    bind_par[5].buffer_type = MYSQL_TYPE_STRING;
    bind_par[5].buffer = (char*) new_usr->name.c_str();
    bind_par[5].buffer_length = new_usr->name.size() + 1;

    // lastname
    bind_par[6].buffer_type = MYSQL_TYPE_STRING;
    bind_par[6].buffer = (char*) new_usr->lastname.c_str();
    bind_par[6].buffer_length = new_usr->lastname.size() + 1;

    //date_of_birth
    char buf[32] = {0};
    strftime(buf, 32, "%Y-%m-%d", &(new_usr->date_of_birth));
    bind_par[7].buffer_type = MYSQL_TYPE_STRING;
    bind_par[7].buffer = buf;
    bind_par[7].buffer_length = strlen(buf);


    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: insert_user. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: insert_user. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: insert_user. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: insert_user. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }

    spdlog::info("db_driver.cpp - fnc: insert_user. New user added successfully.");
    mysql_stmt_close(stmt);
    return 0;
}

int db_driver::insert_measurement(measurement *m) {
    int par_cnt = 15;
    MYSQL_BIND bind_par[par_cnt];

    memset(bind_par, 0, sizeof(MYSQL_BIND) * par_cnt);

    std::string sql = "INSERT INTO cs_measurements(measurement_number, measuree_id, measurer_id, measuring_length,"
                      " measuring_start, measuring_end, description, value_length, value_count, m_values, x_val_median,"
                      " x_val_avg, all_val_median, all_val_avg, measurement_part)"
                      " VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

    // measurement_number
    bind_par[0].buffer_type = MYSQL_TYPE_LONG;
    bind_par[0].buffer = &(m->measurement_number);

    // measuree_id
    bind_par[1].buffer_type = MYSQL_TYPE_LONG;
    bind_par[1].buffer = &(m->measuree_id);

    // measurer_id
    bind_par[2].buffer_type = MYSQL_TYPE_LONG;
    bind_par[2].buffer = &(m->measurer_id);

    // measuring_length
    bind_par[3].buffer_type = MYSQL_TYPE_LONG;
    bind_par[3].buffer = &(m->measuring_length);

    // measuring_start
    char date_start[32] = {0};
    strftime(date_start, 32, "%Y-%m-%d %H:%M:%S", &(m->measuring_start));
    bind_par[4].buffer_type = MYSQL_TYPE_STRING;
    bind_par[4].buffer = date_start;
    bind_par[4].buffer_length = strlen(date_start);

    // measuring_end (FIXME ran into problems on this part when measuring_end wasn't suplied)
    char date_end[32] = {0};
    strftime(date_end, 32, "%Y-%m-%d %H:%M:%S", &(m->measuring_end));
    bind_par[5].buffer_type = MYSQL_TYPE_STRING;
    bind_par[5].buffer = date_end;
    bind_par[5].buffer_length = strlen(date_end);

    // description
    bind_par[6].buffer_type = MYSQL_TYPE_STRING;
    bind_par[6].buffer = (char*) m->description.c_str();
    bind_par[6].buffer_length = m->description.size() + 1;

    // value_length
    bind_par[7].buffer_type = MYSQL_TYPE_TINY;
    bind_par[7].buffer = &(m->value_length);

    // value_count
    bind_par[8].buffer_type = MYSQL_TYPE_LONG;
    bind_par[8].buffer = &(m->value_count);

    // rfid_serial
    bind_par[9].buffer_type = MYSQL_TYPE_MEDIUM_BLOB;
    bind_par[9].buffer = m->values.data();
    bind_par[9].buffer_length = m->values.size() * sizeof(double); // n * s

    // x_val_median
    bind_par[10].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_par[10].buffer = &(m->x_val_med);
    
    // x_val_avg
    bind_par[11].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_par[11].buffer = &(m->x_val_avg);

    // all_val_median
    bind_par[12].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_par[12].buffer = &(m->all_val_med);

    // all_val_avg
    bind_par[13].buffer_type = MYSQL_TYPE_DOUBLE;
    bind_par[13].buffer = &(m->all_val_avg);

    // measurement_part
    bind_par[14].buffer_type = MYSQL_TYPE_LONG;
    bind_par[14].buffer = &(m->measurement_part);

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (stmt == NULL) {
        spdlog::error("db_driver.cpp - fnc: insert_measurement. Failed to init prepared statement.");
        return 1;
    }
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("db_driver.cpp - fnc: insert_measurement. Failed to prepare prepared statement.");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_bind_param(stmt, bind_par)) {
        spdlog::error("db_driver.cpp - fnc: insert_measurement. Failed to bind parameters to prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("db_driver.cpp - fnc: insert_measurement. Failed to execute prepared statement");
        this->handle_stmt_error(stmt);
        return 1;
    }

    spdlog::info("db_driver.cpp - fnc: insert_measurement. Measurement added successfully.");
    mysql_stmt_close(stmt);
    return 0;
}

void db_driver::mysql_time_to_tm(MYSQL_TIME* src, std::tm *dest) {
    // https://stackoverflow.com/questions/421234/convert-mysql-time-data-type-to-char-or-c-string
    char buf[32] = {0};
    sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d", src->year, src->month, src->day, src->hour, src->minute, src->second);
    strptime(buf, "%Y-%m-%d %H:%M:%S", dest);

    // for better performance try: dest->tm_year = src->year; ...
}

void db_driver::handle_stmt_error(MYSQL_STMT* stmt) {
    spdlog::error("db_driver.cpp - Prepared Statement Error ({0}): [{1}] \"{2}\"", 
        mysql_stmt_errno(stmt),
        mysql_stmt_sqlstate(stmt),
        mysql_stmt_error(stmt));

    mysql_stmt_close(stmt);
}

// int db_driver::prprd_stmt_executor(const char *sql, const char* fnc_name,
//     MYSQL_BIND *par, int par_cnt, MYSQL_BIND *res, int res_cnt)
// {

// }

/*****
 *  tests 
*****/

static void show_stmt_error(MYSQL_STMT *stmt)
{
  printf("Error(%d) [%s] \"%s\"", mysql_stmt_errno(stmt),
                                  mysql_stmt_sqlstate(stmt),
                                  mysql_stmt_error(stmt));
}

static char text_test[64] = "insert from c";
static uint8_t binary_data_test[] = {0x10, 0x50, 0xFF, 0x07, 0xAB, 0xB7, 0xBC, 0xCB, 0xCC, 0x77};
static int int_test = 5;
static float float_test = 8.00719;

void db_driver::test_insert_with_binary() {
    int bind_count = 5;
    MYSQL_BIND bind[bind_count];
    std::string sql("INSERT INTO test(text_test, int_test, float_test, binary_data, blob_data) VALUES (?, ?, ?, ?, ?)");
    // std::string sql("INSERT INTO test(text_test, int_test) VALUES (?,?)");

    memset(bind, 0, sizeof(MYSQL_BIND) * bind_count);

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("Failed prepared statment.");
        show_stmt_error(stmt);
        return;
    }

    unsigned long bin_count = 10;

    bind[0].buffer_type = MYSQL_TYPE_STRING; // string works but VARCHAR doesnt, i guess it expects something more
    //bind[0].buffer_type = MYSQL_TYPE_VARCHAR;
    bind[0].buffer = text_test;
    bind[0].buffer_length = strlen(text_test);

    bind[1].buffer_type = MYSQL_TYPE_LONG;
    bind[1].buffer = &int_test;

    bind[2].buffer_type = MYSQL_TYPE_FLOAT;
    bind[2].buffer = &float_test;

    bind[3].buffer_type = MYSQL_TYPE_STRING;
    bind[3].buffer = binary_data_test;
    bind[3].length = &bin_count;

    bind[4].buffer_type = MYSQL_TYPE_MEDIUM_BLOB;
    bind[4].buffer = binary_data_test;
    bind[4].length = &bin_count;

    // int array_size = 5;
    // mysql_stmt_attr_set(stmt, STMT_ATTR_ARRAY_SIZE, &array_size);

    mysql_stmt_bind_param(stmt, bind);

    if (mysql_stmt_execute(stmt))
        show_stmt_error(stmt);

    mysql_stmt_close(stmt);
}

void db_driver::test_select_with_binary() {
    int params = 1;
    int results = 5;
    MYSQL_BIND bind_param[params];
    MYSQL_BIND bind_res [results];
    unsigned long length[results] = {0};
    my_bool is_null     [results] = {0};
    my_bool error       [results] = {0};

    //std::string sql("SELECT text_test, int_test, float_test, binary_data, blob_data FROM test WHERE id = ?");
    // std::string sql("SELECT text_test, int_test, float_test, binary_data, blob_data FROM test WHERE text_test LIKE ?");
    std::string sql("SELECT text_test, int_test, float_test, binary_data, blob_data FROM test WHERE blob_data = ?");
    // std::string sql("INSERT INTO test(text_test, int_test) VALUES (?,?)");

    memset(bind_param, 0, sizeof(MYSQL_BIND) * params);
    memset(bind_res, 0, sizeof(MYSQL_BIND) * results);

    MYSQL_STMT *stmt = mysql_stmt_init(conn.get());
    if (mysql_stmt_prepare(stmt, sql.c_str(), strlen(sql.c_str()))) {
        spdlog::error("Failed prepared statment.");
        show_stmt_error(stmt);
        return;
    }

    // long searched_id = 5;
    // char* search_string = "%%from%";
    unsigned long bin_count = 10;

    float f_in;
    int i_in;
    char t_in[128];
    uint8_t bin_in[128];
    uint8_t blob_in[128];

    // string where
    // bind_param[0].buffer_type = MYSQL_TYPE_STRING; // string works but VARCHAR doesnt, i guess it expects something more
    // bind_param[0].buffer = search_string;
    // bind_param[0].buffer_length = strlen(search_string);

    // int (long) where
    // bind_param[0].buffer_type = MYSQL_TYPE_LONG;
    // bind_param[0].buffer = &searched_id;

    bind_param[0].buffer_type = MYSQL_TYPE_MEDIUM_BLOB;
    bind_param[0].buffer = binary_data_test;
    bind_param[0].length = &bin_count;

    bind_res[0].buffer_type = MYSQL_TYPE_STRING; // string works but VARCHAR doesnt, i guess it expects something more
    bind_res[0].buffer = t_in;
    bind_res[0].buffer_length = 128;
    bind_res[0].is_null = &is_null  [0];
    bind_res[0].length = &length    [0];
    bind_res[0].error = &error      [0];

    bind_res[1].buffer_type = MYSQL_TYPE_LONG;
    bind_res[1].buffer = &i_in;
    bind_res[1].is_null = &is_null  [1];
    bind_res[1].length = &length    [1];
    bind_res[1].error = &error      [1];

    bind_res[2].buffer_type = MYSQL_TYPE_FLOAT;
    bind_res[2].buffer = &f_in;
    bind_res[2].is_null = &is_null  [2];
    bind_res[2].length = &length    [2];
    bind_res[2].error = &error      [2];

    bind_res[3].buffer_type = MYSQL_TYPE_STRING;
    bind_res[3].buffer = bin_in;
    bind_res[3].buffer_length = 128;
    bind_res[3].is_null = &is_null  [3];
    bind_res[3].length = &length    [3];
    bind_res[3].error = &error      [3];

    bind_res[4].buffer_type = MYSQL_TYPE_MEDIUM_BLOB;
    bind_res[4].buffer = blob_in;
    bind_res[4].buffer_length = 128;
    bind_res[4].is_null = &is_null  [4];
    bind_res[4].length = &length    [4];
    bind_res[4].error = &error      [4];

    // int array_size = 5;
    // mysql_stmt_attr_set(stmt, STMT_ATTR_ARRAY_SIZE, &array_size);

    if (mysql_stmt_bind_param(stmt, bind_param)) {
        spdlog::error("Failed to bind parameters");
        return;
    }
    if (mysql_stmt_bind_result(stmt, bind_res)) {
        spdlog::error("Failed to bind results");
        return;
    }
    if (mysql_stmt_execute(stmt)) {
        spdlog::error("Failed to execute statement");
        show_stmt_error(stmt);
        return;
    }
    if (mysql_stmt_store_result(stmt)) {
        spdlog::error("Failed to store result");
        show_stmt_error(stmt);
        return;
    }

    //spdlog::info("here");
    // printf("rv: %d, %d, %d\n", rv, MYSQL_NO_DATA, MYSQL_DATA_TRUNCATED);
    while (1) {
        int res = mysql_stmt_fetch(stmt);
        if (res == 1 || res == MYSQL_NO_DATA) {
            spdlog::info("breaking with {0}", res);
            break;
        }

        spdlog::info("Fetched result");

        for (int i = 0; i < results; i++) {
            printf("result flags: is_null: %c\terror: %c\tlength: %ld\n", is_null[i], error[i], length[i]);
            if (is_null[i]) {
                printf("value %d is null\n", i);
            }
            if (error[i]) {
                printf("error in val %d : %d\n", i, error[i]);
            }
            // printf("%s\n", t_in);
        }

        printf("%s\n", t_in);
        printf("%d\n", i_in);
        printf("%f\n", f_in);

        // test blob
        long unsigned int len = bin_count;
        bool correct = true;
        if (bin_count != length[results - 1]) {
            printf("blob size mismatch\n");
            len = length[results - 1] < len ? length[results - 1] : len;
        }
        for (long unsigned int i = 0; i < len; i++) {
            if (binary_data_test[i] != blob_in[i]) {
                printf("blob data aren't correct on byte %ld\n", i);
                correct = false;
            }
        }

        if (correct) {
            printf("YEEEEY! blob data match\n");
        } else {
            printf("OUCH! blob data dont match\n");
        }

        // test binary
        long unsigned int len2 = bin_count;
        bool correct2 = true;
        if (bin_count != length[results - 2]) {
            printf("binary size mismatch, but its expected. Binary is padded with 0 from right\n");
            len2 = length[results - 2] < len2 ? length[results - 2] : len2;
        }
        for (long unsigned int i = 0; i < len2; i++) {
            if (binary_data_test[i] != blob_in[i]) {
                printf("binary data aren't correct on byte %ld\n", i);
                correct2 = false;
            }
        }

        if (correct2) {
            printf("YEEEEY! binary data match\n");
        } else {
            printf("OUCH! binary data dont match\n");
        }

    }

    mysql_stmt_close(stmt);
}

/**
 * @brief This was an initial test without using prepared statements, this shouldn't be used anywhere
 * 
 * @return int 
 */
int db_driver::test() {
   // Initialize Connection
   MYSQL *conn;
   if (!(conn = mysql_init(0)))
   {
      fprintf(stderr, "unable to initialize connection struct\n");
      exit(1);
   }

   // Connect to the database
   if (!mysql_real_connect(
         conn,                 // Connection
         "localhost", // Host
         "pi",            // User account
         "raspberry",   // User password
         "control_system",               // Default database
         3306,                 // Port number
         NULL,                 // Path to socket file
         0                     // Additional options
      ))
   {
    printf("Failed to open mysql connection.\n");
      // Report the failed-connection error & close the handle
      fprintf(stderr, "Error connecting to Server: %s\n", mysql_error(conn));
      mysql_close(conn);
      exit(1);
   }

   // Use the Connection
   // ...
   const char* sql = "SELECT * FROM test";
   if (mysql_real_query(conn, sql, strlen(sql))) {
      printf("Failed %s\n", sql);
      return 1;
   }

   MYSQL_RES* res = mysql_store_result(conn);
   if (!res) {
      printf("RESULT SET is null/ empty\n");
      return 1;
   }

   MYSQL_ROW row;
   unsigned int num_fields;
   unsigned int i;

   num_fields = mysql_num_fields(res);
   while ((row = mysql_fetch_row(res))) {
    unsigned long *lengths;
       lengths = mysql_fetch_lengths(res);
       for(i = 0; i < num_fields; i++)
       {
          printf("[%.*s] ", (int) lengths[i],
                 row[i] ? row[i] : "NULL");
       }
       printf("\n");
   }
   /*
   int cols = mysql_field_count(conn);
   printf("%s returns %d cols\n", sql, cols);
   */

   mysql_free_result(res);

   // Close the Connection
   mysql_close(conn);

   return 0;
}