#include "user_workspace.h"

#include <memory.h>
#include <spdlog/spdlog.h>
#include "localisation.h"


#define X_SAMPLES 5

/*********************************************************************************************************************/
/*       Measurements function definitions                                                                           */
/*********************************************************************************************************************/

const char* measurement_header::get_header_label() {
    char buf[64] = {0};
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->measuring_start));
    this->label = std::to_string(this->measurement_number) + " - " + std::string(buf);

    return this->label.c_str();
}

void measurement::log_measurement_to_debug() {
    spdlog::debug("measurement - ID: {0}", this->id);
    // spdlog::debug("measurement - status: {0}", this->status);
    spdlog::debug("measurement - measurement_number: {0}", this->measurement_number);
    spdlog::debug("measurement - measuree_id: {0}", this->measuree_id);
    spdlog::debug("measurement - measurer_id: {0}", this->measurer_id);
    spdlog::debug("measurement - measuring_length: {0}", this->measuring_length);
    char buf[64] = {0};
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->measuring_start));
    spdlog::debug("measurement - measuring start: {0}", buf);
    memset(buf, 0, 64);
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->measuring_end));
    spdlog::debug("measurement - measuring end: {0}", buf);
    spdlog::debug("measurement - description: {0}", this->description.c_str());
    spdlog::debug("measurement - value_length: {0}", this->value_length);
    spdlog::debug("measurement - value_count: {0}", this->value_count);
    spdlog::debug("measurement - raw_values size: {0}", this->values.size());
    spdlog::debug("measurement - x_val_med: {0}", this->x_val_med);
    spdlog::debug("measurement - x_val_avg: {0}", this->x_val_avg);
    spdlog::debug("measurement - all_val_med: {0}", this->all_val_med);
    spdlog::debug("measurement - all_val_avg: {0}", this->all_val_avg);
    memset(buf, 0, 64);
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->created_at));
    spdlog::debug("measurement - created at: {0}", buf);
}

void measurement::init_value_strings() {
    // Mass must be created as G. The hx sensor library atually returns measured double in Grams!
    HX711::Mass x_med(x_val_med, HX711::Mass::Unit::G);
    HX711::Mass x_avg(x_val_avg, HX711::Mass::Unit::G);
    HX711::Mass all_med(all_val_med, HX711::Mass::Unit::G);
    HX711::Mass all_avg(all_val_avg, HX711::Mass::Unit::G);

    /*
    spdlog::info("Measured value no par: {0}", x_med.toString().c_str());
    spdlog::info("Measured value MG: {0}", x_med.toString(HX711::Mass::Unit::MG).c_str());
    spdlog::info("Measured value G: {0}", x_med.toString(HX711::Mass::Unit::G).c_str());
    spdlog::info("Measured value KG: {0}", x_med.toString(HX711::Mass::Unit::KG).c_str());
    */

    this->x_val_med_str = x_med.toString(current_unit);
    this->x_val_avg_str = x_avg.toString(current_unit);
    this->all_val_med_str = all_med.toString(current_unit);
    this->all_val_avg_str = all_avg.toString(current_unit);
}

void measurement::init_start_end_strings() {
    char buf[64] = {0};
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->measuring_start));
    // FIXME - is first or second version better?
    // this->start_str = buf;
    this->start_str = std::string(buf);

    memset(buf, 0, 64);
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->measuring_end));
    // this->end_str = buf;
    this->end_str = std::string(buf);

    memset(buf, 0, 64);
    if (this->continuous) {
        // continuous measuring is minute at least, up to days, so length is in s
        sprintf(buf, "%ld s", std::chrono::seconds(this->measuring_length).count());
    } else {
        // regular measuring is expected to take around a minute max
        sprintf(buf, "%ld ms", std::chrono::duration_cast<std::chrono::milliseconds>
            (std::chrono::microseconds(this->measuring_length)).count());
    }
    // this->length_str = buf;
    this->length_str = std::string(buf);
}

void measurement::init_convinience_vars(const char* measurer_uname, const char* measuree_uname) {
    this->init_value_strings();
    this->init_start_end_strings();
    this->measuree_uname = measuree_uname;
    this->measurer_uname = measurer_uname;
}

int measurement::compare(measurement* m) {
    spdlog::debug("Comparing:");
    this->log_measurement_to_debug();
    spdlog::debug("to:");
    m->log_measurement_to_debug();
    spdlog::info("Comparing start");
    
    if (this->measurement_number != m->measurement_number) {
        spdlog::debug("measurement_number diff: {0} - {1}", this->measurement_number, m->measurement_number);
    }
    if (this->measuree_id != m->measuree_id) {
        spdlog::debug("measuree_id diff: {0} - {1}", this->measuree_id, m->measuree_id);
    }
    if (this->measurer_id != m->measurer_id) {
        spdlog::debug("measurer_id diff: {0} - {1}", this->measurer_id, m->measurer_id);
    }
    if (this->measuring_length != m->measuring_length) {
        spdlog::debug("measuring_length diff: {0} - {1}", this->measuring_length, m->measuring_length);
    }
    if (this->x_val_med != m->x_val_med) {
        spdlog::debug("x_val_med diff: {0} - {1}", this->x_val_med, m->x_val_med);
    }
    if (this->x_val_avg != m->x_val_avg) {
        spdlog::debug("x_val_avg diff: {0} - {1}", this->x_val_avg, m->x_val_avg);
    }
    if (this->all_val_med != m->all_val_med) {
        spdlog::debug("all_val_med diff: {0} - {1}", this->all_val_med, m->all_val_med);
    }
    if (this->all_val_avg != m->all_val_avg) {
        spdlog::debug("all_val_avg diff: {0} - {1}", this->all_val_avg, m->all_val_avg);
    }
    if (this->values.size() != m->values.size()) {
        spdlog::debug("values.size() diff: {0} - {1}", this->values.size(), m->values.size());
    }

    size_t len = this->values.size() > m->values.size() ? m->values.size() : this->values.size();
    for (size_t i = 0; i < len; i++) {
        if (this->values[i] != m->values[i]) {
            spdlog::error("values (binary) start to differ at index {0}", i);
            break;
        }
    }

    spdlog::info("Comparing end");
    return 0;
}

void measurement::init_simple_measuring(measurement *m, std::vector<double> &values) {
    // this is called on every update - not logging even to debug
    // spdlog::debug("user_workspace.cpp - Initing simple measurements");
    m->values = values;
    m->value_count = values.size();
    m->value_length = sizeof(double);

    measurement::calc_measuring_results(m);
    m->init_value_strings();
}

void measurement::init_measuring(measurement *m, user_workspace *uspace, std::vector<double> &values,
    std::tm *start_tm, std::tm *end_tm, unsigned long m_length)
{
    spdlog::debug("user_workspace.cpp - Initing basic measuring parameters");

    m->values = values;
    m->measuree_id = uspace->has_subuser() ? uspace->subuser->id : uspace->get_user_id();
    m->measurer_id = uspace->get_user_id();
    m->measurement_number = uspace->has_subuser() ? uspace->subuser->measure_count + 1 :
                                                    uspace->get_user()->measure_count + 1;
    
    // dereferencing like this should copy
    m->measuring_start = *start_tm;
    m->measuring_end = *end_tm;
    m->measuring_length = m_length;
    m->value_count = values.size();
    m->value_length = sizeof(double);

    measurement::calc_measuring_results(m);
}

void measurement::calc_measuring_results(measurement *m) {
    // called often by init_simple_measuring(...); disabling debug
    // spdlog::debug("user_workspace.cpp - Initing calculated measuring parameters");
    std::pair<double, double> x_vals = measurement::calc_every_x_values_stats(m);
    m->x_val_med = x_vals.first;
    m->x_val_avg = x_vals.second;

    std::pair<double, double> all_vals = measurement::calc_all_values_stats(m);
    m->all_val_med = all_vals.first;
    m->all_val_avg = all_vals.second;
}

void measurement::prepare_continuous_to_picked(std::vector<measurement> &ms, measurement* picked) {
    if (ms.size() <= 0) return; // this shouldn't occure, but just to be safe

    *picked = ms[0]; // copy first element to picked

    picked->measuring_end = ms[ms.size() - 1].measuring_end;
    picked->cont_measurements = ms.size();
    picked->continuous = true;

    unsigned long total_length = 0;
    double xm_sum = 0.0, xa_sum = 0.0, am_sum = 0.0, aa_sum = 0.0;

    for (size_t i = 0; i < ms.size(); i++) {
        // measuring length is in us, continuous length is assumed to be minutes, hours or even days -> stored in seconds
        total_length += ms[i].measuring_length / 1000000;
        xm_sum += ms[i].x_val_med;
        xa_sum += ms[i].x_val_avg;
        am_sum += ms[i].all_val_med;
        aa_sum += ms[i].all_val_avg;
    }

    picked->measuring_length = total_length;
    // use avarages of all stats as result to show
    picked->x_val_med = xm_sum / ms.size();
    picked->x_val_avg = xa_sum / ms.size();
    picked->all_val_med = am_sum / ms.size();
    picked->all_val_avg = aa_sum / ms.size();
}

std::pair<double, double> measurement::calc_every_x_values_stats(measurement *m) {
    size_t it = 0, inc = 0;
    std::vector<double> vect = m->values; // copy to not modify original vector order
    std::vector<double> meds;
    std::vector<double> avgs;

    while (it < vect.size()) {
        if ((it + X_SAMPLES) < vect.size())
            inc = X_SAMPLES;
        else
            inc = vect.size() - it;

        auto it_start = vect.begin() + it;
        auto it_end = vect.begin() + it + inc;

        std::vector<double> tmp(it_start, it_end);

        meds.push_back(HX711::Utility::median(&tmp));
        avgs.push_back(HX711::Utility::average(&tmp));

        it += inc;
    }

    return std::pair<double, double>(HX711::Utility::median(&meds), HX711::Utility::average(&avgs));
}

std::pair<double, double> measurement::calc_all_values_stats(measurement *m) {
    std::vector<double> vect = m->values; // copy to not modify original vector order
    return std::pair<double, double>(HX711::Utility::median(&vect), HX711::Utility::average(&vect));
}


/*********************************************************************************************************************/
/*       User container function definitions                                                                         */
/*********************************************************************************************************************/

user_cont::user_cont(unsigned long id, uint8_t *rfid_ser_bin, uint8_t rfid_ser_len, std::string username,
    std::string name, std::string lastname, std::string description, uint8_t role, std::tm reg_date,
    std::tm last_login_date, std::tm created_at_date, std::tm date_of_birth)
{
    this->id = id;
    this->username = username;
    this->rfid_ser_len = rfid_ser_len < 10 ? rfid_ser_len : 10;
    memcpy(this->rfid_serial_bin, rfid_ser_bin, this->rfid_ser_len);
    this->name = name;
    this->lastname = lastname;
    this->description = description;
    this->role = role;
    this->reg_date = reg_date;
    this->last_login_date = last_login_date;
    this->created_at_date = created_at_date;
    this->date_of_birth = date_of_birth;
}

void user_cont::log_user_to_debug() {
    spdlog::debug("user_cont - User ID: {0}", this->id);
    spdlog::debug("user_cont - User RFID: {0:x} {1:x} {2:x} {3:x} {4:x} {5:x} {6:x} {7:x} {8:x} {9:x}", 
        this->rfid_serial_bin[0], this->rfid_serial_bin[1], this->rfid_serial_bin[2], this->rfid_serial_bin[3],
        this->rfid_serial_bin[4], this->rfid_serial_bin[5], this->rfid_serial_bin[6], this->rfid_serial_bin[7],
        this->rfid_serial_bin[8], this->rfid_serial_bin[9]);
    spdlog::debug("user_cont - User rfid_ser_len: {0}", this->rfid_ser_len);
    spdlog::debug("user_cont - User username: {0}", this->username);
    spdlog::debug("user_cont - User name: {0}", this->name);
    spdlog::debug("user_cont - User lastname: {0}", this->lastname);
    spdlog::debug("user_cont - User description: {0}", this->description);
    spdlog::debug("user_cont - User role: {0}", this->role);

    char buf[64] = {0};
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->reg_date));
    spdlog::debug("user_cont - User reg_date: {0}", buf);
    memset(buf, 0, 64);
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->last_login_date));
    spdlog::debug("user_cont - User last_login_date: {0}", buf);
    memset(buf, 0, 64);
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->created_at_date));
    spdlog::debug("user_cont - User created_at_date: {0}", buf);
    memset(buf, 0, 64);
    strftime(buf, 64, "%Y-%m-%d %H:%M:%S", &(this->date_of_birth));
    spdlog::debug("user_cont - User date_of_birth: {0}", buf);
}


/*********************************************************************************************************************/
/*       User_workspace function definitions                                                                         */
/*********************************************************************************************************************/

user_workspace::user_workspace(user_cont* user) {
    this->user.reset(user);
    this->measured.reset(new measurement());
    this->picked.reset(new measurement());
}

measurement* user_workspace::get_last_hx_measuring() {
    return this->measured.get();
}

measurement* user_workspace::get_picked_hx_measuring() {
    return this->picked.get();
}

int user_workspace::compare_rfid_serials(rfid_reader::rfid_tag *tag) {
    if (tag->serial_size != this->user->rfid_ser_len)
        return 1;

    for (int i = 0; i < tag->serial_size; i++) {
        if (this->user->rfid_serial_bin[i] != tag->serial[i])
            return 1;
    }

    return 0;
}


bool user_workspace::has_user() {
    return this->user.get() != nullptr;
}

bool user_workspace::has_subuser() {
    return this->subuser.get() != nullptr;
}

std::pair<uint8_t*, uint8_t> user_workspace::get_rfid_serial() {
    return std::pair<uint8_t*, uint8_t>(this->user->rfid_serial_bin, this->user->rfid_ser_len);
}

unsigned long user_workspace::get_user_id() {
    return this->user->id;
}

user_cont* user_workspace::get_user() {
    return this->user.get();
}

std::string& user_workspace::get_username() {
    return this->user->username;
}

std::string& user_workspace::get_name() {
    return this->user->name;
}

std::string& user_workspace::get_lastname() {
    return this->user->lastname;
}
std::string& user_workspace::get_description() {
    return this->user->description;
}

uint8_t user_workspace::get_role() {
    return this->user->role;
}

const char* user_workspace::get_role_string() {
    switch (this->user->role) {
        case 0:
            return get_localized_text("GT_USER_ROLE_PAC");
        case 1:
            return get_localized_text("GT_USER_ROLE_EMP");
        case 2:
            return get_localized_text("GT_USER_ROLE_ADM");
        
        default:
            return get_localized_text("GT_USER_ROLE_ERR");
    }
}

std::vector<std::unique_ptr<measurement_header>>& user_workspace::get_user_measur_headers() {
    if (this->subuser.get() != nullptr) {
        return this->subuser->measur_headers;
    } else {
        return this->user->measur_headers;
    }
}

user_cont* user_workspace::get_measured_user() {
    if (this->subuser.get() != nullptr) {
        return this->subuser.get();
    } else {
        return this->user.get();
    }   
}

unsigned long user_workspace::get_measured_id() {
     if (this->subuser.get() != nullptr) {
        return this->subuser->id;
    } else {
        return this->user->id;
    }
}

std::string& user_workspace::get_measured_username() {
     if (this->subuser.get() != nullptr) {
        return this->subuser->username;
    } else {
        return this->user->username;
    }
}

void user_workspace::init_measured(std::vector<double> values, std::tm *start_tm, std::tm *end_tm, unsigned long m_length) {
    spdlog::debug("user_workspace.cpp - Initing basic measuring parameters");
    this->measured.reset(new measurement());

    measurement::init_measuring(this->measured.get(), this, values, start_tm, end_tm, m_length);
}